#!/usr/bin/env python

#   Features to be added later:
#proper splice site modeler (perhaps optimizing splice site score vs. difference from expected length), maybe MAYBE taking
#into account non-conserved exon phase.



import argparse
from HAPpy import genome_fork as genome
from ABCENTH.FindOrfs import orf_finder,genewisesearch
import sys, os

def isyes(x):
    if x in [True,'True','true','TRUE','TrUe','T','t','oui','si','y','Y','Yes','YES','yes','oh my YESSSSS!']:
        return True
    else:
        return False


#The heart of gold of this program- the function to define an exon based on the coords and target phase
def exon_finder(tstart,tend,strand,qstart,qend,qlen,qstartphase,qendphase,seqdict,seqname,
                max_offset = 30, is_start = False, is_stop = False, nevermind_atg = False,
                cluster = None,exon_number = None,log_file = open(os.devnull, 'w'),full_pseudoexon_search = True,
                exon_info_dict = False):
    """"finds exons with ORFs based on the requested parameters. Note that it is expected that tstart < tend, \
    even for minus strand features, so these might need to be reversed if coming from say tblastn. \
    Also coords are expected as 1-based (as output from blast), and are \
    converted internally to 0 based. It is also expected that the hit itself doesn't contain stop codons."""
    start = None
    end = None
    pseudo = False
    max_coord = len(seqdict[seqname]) - 1
    if strand == "+":
        phasestart_offset = (3 - qstartphase) % 3
        phasestop_offset = qendphase
        start_match_offset, stop_match_offset = 3 * (qstart - 1),3 * (qlen - qend)
    elif strand == '-':
        phasestart_offset = qendphase
        phasestop_offset = (3 - qstartphase) % 3
        start_match_offset, stop_match_offset = 3 * (qlen - qend), 3 * qstart
    ideal_start = tstart - phasestart_offset - start_match_offset
    ideal_end = tend + stop_match_offset + phasestop_offset
    pseudo_start = tstart - phasestart_offset
    pseudo_end = tend + phasestop_offset
    gc_start, gc_end = None, None
    for offset in range(0,max_offset + 3,3):
        if start:
            break
        for direction in [1,-1]:
            test_start = ideal_start - offset * direction
            test_seq = genome.Sequence(seqdict[seqname][test_start-1 + phasestart_offset:tend])
            if strand == "-":
                test_seq = test_seq.reverse_compliment()
            if not test_seq.translate():
                continue
            elif is_stop and strand == "-":
                if ideal_start - 1 < test_start < pseudo_start:
                    pseudo_start = test_start
                lastcodon = seqdict[seqname][test_start - 1:test_start + 2]
                if lastcodon.upper() in ['TTA','TCA','CTA'] and test_seq.translate().count('*') == 1:
                    start = test_start
                    break
            elif not "*" in test_seq.translate():
                if ideal_start - 1 < test_start < pseudo_start:
                    pseudo_start = test_start
                if is_start and strand == '+':
                    if nevermind_atg:
                        start = test_start
                        break
                    else:
                        firstcodon = seqdict[seqname][test_start - 1:test_start + 2]
                        if firstcodon.upper() == "ATG":
                            start = test_start
                            break
                else:
                    splicesite = seqdict[seqname][test_start-3:test_start-1]
                    if (strand == '+' and splicesite.upper() == "AG") or (strand == '-' and splicesite.upper() == "AC"):
                        start = test_start
                        break
                    elif strand == '-' and splicesite.upper() == "GC" and not gc_start:
                        gc_start = test_start
    if not start:
        if gc_start:
            start = gc_start
        else:
            pseudo = "P"
            start = pseudo_start
    for offset in range(0,max_offset + 3,3):
        if end:
            break
        for direction in [-1,1]:
            test_end = ideal_end - offset * direction
            if test_end - start < 3:
                break
            test_seq = genome.Sequence(seqdict[seqname][start - 1 + phasestart_offset:test_end - phasestop_offset])
            if strand == "-":
                test_seq = test_seq.reverse_compliment()
            if not test_seq.translate():
                continue
            elif is_stop and strand == "+":
                if ideal_end + 1 > test_end > pseudo_end:
                    pseudo_end = test_end
                lastcodon = seqdict[seqname][test_end - 3:test_end]
                if lastcodon.upper() in ['TAA','TGA','TAG'] and test_seq.translate().count('*') == 1:
                    end = test_end
                    break
            elif not "*" in test_seq.translate() or (is_stop and not "*" in test_seq.translate()[:-1]):
                if ideal_end + 1 > test_end > pseudo_end:
                    pseudo_end = test_end
                if is_start and strand == '-':
                    if nevermind_atg:
                        end = test_end
                        break
                    else:
                        firstcodon = seqdict[seqname][test_end - 3:test_end]
                        if firstcodon.upper() == "CAT":
                            end = test_end
                            break
                else:
                    splicesite = seqdict[seqname][test_end:test_end + 2]
                    if (strand == '+' and splicesite.upper() == "GT") or (strand == '-' and splicesite.upper() == "CT"):
                        end = test_end
                        break
                    elif strand == "+" and splicesite.upper() == "GC" and not gc_end:
                        gc_end = test_end
    if not end:
        if gc_end:
            end = gc_end
        else:
            pseudo = "P"
            end = pseudo_end
    start = max([1,start])
    end = min([end,max_coord])
    if pseudo and full_pseudoexon_search and cluster != None and exon_number != None and exon_info_dict:
        gwexons = genewisesearch(seqdict[seqname],qstartphase,qendphase,strand,
                    exon_info_dict[str(cluster) + ':' + str(exon_number)][6], 
                    search_coords = [ideal_start - 3 - max_offset,ideal_end + 3 + max_offset],
                    seqname = seqname,log_file=log_file)
        if gwexons != []:
            return gwexons
    return [[start,end,pseudo]]

#exon_finder(4179626,4179670,"-",1,40,40,0,0,cra.genome_sequence,'Chromosome3') #debug>>> exon_finder(4181323,4181442,"-",1,40,40,2,2,cra.genome_sequence,'Chromosome3') #debug

def exontuples2gff(exontuple_list,strand,feature_name,locus_name):
    outlines = []
    problems = []
    is_pseudo = False
    for exontuple in exontuple_list:
        outlines.append('\t'.join([locus_name,'ABCENTH','CDS',str(min(exontuple[:2])),str(max(exontuple[:2])),'.',strand,
                                  '.','gene_id ' + feature_name + ';transcript_id ' + feature_name + '-RA']))
        if exontuple[2]:
            is_pseudo = True
            problems.extend(list(exontuple[2]))
    if is_pseudo:
        problems = "".join(list(set(problems)))
        return '\n'.join(outlines).replace(';',problems + ';').replace('-RA',problems + '-RA')
    else:
        return '\n'.join(outlines)



#Defines a few operations that will need to be executed in different parts of the following code. NOT independent functions!
def recover_missing_exon(strand,cluster,exon_numbers,includes_start,includes_stop, search_coords, log_file,args,exon_info_dict,
                        target_genome,locus,tstart,last_strand,working_annotation,full_pseudoexon_search ):
    log_file.write('Exons skipped, attempting to recover.\n')
    # V- I think I fixed it
    #needs to be fixed to adjust phases in case not all exons are found
    found_coords_list = []
    adjust_next_phase = 0
    for exon_number in exon_numbers:
        if includes_start and exon_number == exon_numbers[0]:
            is_start,is_stop = True,False
        elif includes_stop and exon_number == exon_numbers[-1]:
            is_start,is_stop = False,True
        else:
            is_start,is_stop = False,False
        if "query_exon_info_table" in args:
            missing_exon_info = exon_info_dict[str(cluster) + ':' + str(exon_number)]
            missing_coords = orf_finder(target_genome.genome_sequence[locus],int(missing_exon_info[3]),
                                                 int(missing_exon_info[4]),strand, int(missing_exon_info[5]),5, search_coords,
                                                 is_stop = is_stop,is_start = is_start,hmm_profile = missing_exon_info[6],
                                                 evalue = args.orf_finder_E, genewise_on_fail = full_pseudoexon_search)
        log_file.write(locus + '\t' + str(tstart) + '\t' + str(cluster) + '\t' + str(exon_number) + '\t' + str(search_coords) + '\t' + str(missing_coords) + '\n')
        if missing_coords != []:
            if found_coords_list != []:
                if ((found_coords_list[-1][1] > missing_coords[0][0] and last_strand == "+") or
                (found_coords_list[-1][0] < missing_coords[-1][1] and last_strand == "-")):
                    adjust_next_phase = ( int(missing_exon_info[3]) - int(missing_exon_info[4]) ) % 3
                    continue
            if last_strand == "+":
                missing_coords[0][0] += adjust_next_phase
            elif last_strand == "-":
                missing_coords[-1][1] -=  adjust_next_phase
            found_coords_list.extend(missing_coords)
            adjust_next_phase = 0
        else:
            adjust_next_phase = (adjust_next_phase + int(missing_exon_info[3]) - int(missing_exon_info[4]) ) % 3
    if adjust_next_phase != 0 and found_coords_list != []:
        if last_strand == "+":
            found_coords_list[-1][1] -= adjust_next_phase % 3
        elif last_strand == "-":
            found_coords_list[-1][0] += adjust_next_phase % 3
    if strand == "-":
        found_coords_list.reverse()
    if found_coords_list != []:
        working_annotation.extend(found_coords_list)
        if len(found_coords_list) == len(exon_numbers):
            return (working_annotation, True)
        else:
            return (working_annotation,"I")
    else:
        return (working_annotation, False)


def last_annotation_almost_complete(last_strand,working_name,last_exon_num,last_tend,last_tstart,tstart,
                                    tend,log_file,args,exon_info_dict,target_genome,locus,working_annotation,
                                    last_startphase,last_num_exons,full_pseudoexon_search):
    if last_strand == '-':
        is_start, is_stop, if_missing = True,False,"N"
        working_annotation, recovered = recover_missing_exon(last_strand,working_name[0].split('coord')[0],range(last_exon_num),
                                        is_start,is_stop,[max((last_tend,last_tstart)),min(max((last_tend,last_tstart)) + 2000,
                                        min((tstart,tend)))],log_file,args,exon_info_dict,target_genome,locus,tstart,last_strand,
                                        working_annotation,full_pseudoexon_search)
    elif last_strand == '+':
        is_start,is_stop, if_missing = False,True,"C"
        working_annotation, recovered = recover_missing_exon(last_strand,working_name[0].split('coord')[0],range(last_exon_num + 1, last_num_exons),
                                        is_start,is_stop,[max((last_tend,last_tstart)),min(max((last_tend,last_tstart)) + 2000,
                                        min((tstart,tend)))],log_file,args,exon_info_dict,target_genome,locus,tstart,last_strand,
                                        working_annotation,full_pseudoexon_search)
    if not recovered:
        if last_strand == '-' and last_startphase != 0:
            working_annotation[-1][1] = working_annotation[-1][1] - ((3 - last_startphase) % 3)
        if working_annotation[-1][2]:
            working_annotation[-1][2] = working_annotation[-1][2] + if_missing
        else:
            working_annotation[-1][2] = if_missing
    log_file.write('writing almost complete annotation "' + working_name[0] + '". \n' )
    print(exontuples2gff(working_annotation, last_strand,working_name[0],locus))

def main():
    parser = argparse.ArgumentParser(description='ABCENTH, Annotation Based on Conserved Exons Noticed Through Homology. Takes a table of \
                                    exon alignments with phase and length information and tries to build good annotations. \
                                    Currently assumes that hits DO NOT contain stop codons.')

    parser.add_argument('--table', help = 'Table of exon alignments. Expected to be in tblastn standard tab delimited format \
                        (-outfmt 6; query,target,blah,blah,blah,blah,qstart,qstop,tstart,tstop,evalue,score) with additional \
                        fields for query exon start phase, query exon end phase, query exon peptide length, query exon number (from 0), and highest query exon number \
                        (e.g. if their are 3 exons, they would be number 0, 1, and 2, so highest query exon number would be 2). \
                        Assumed to contain only non-overlapping hits (use HitTabFilter.py to find best non-overlapping hits)')

    parser.add_argument('--genome', help = 'Genome for which to generate annotations (needed to ID splice sites)')

    parser.add_argument('--sep', default = 'tab', help = 'Hit table field delimiter, accepts "comma" and "tab". Default= "tab"')

    parser.add_argument('--maxintron', type = int, default = 100000, help = 'Maximum lenth of introns. Default = 100000')

    parser.add_argument('--query_exon_info_table', help = 'Information about each query exon (for finding missing exons)')

    parser.add_argument('--full_pseudoexon_search', default = True, type = isyes, 
                        help = 'Experimental (in progress): For truncated exons, performs genewise search \
                        against exon hmm to fill out exon in case of frameshift or stop codon')

    parser.add_argument('--orf_finder_E', type = float, default = 0.05, 
                        help = 'minimum hmmer E value to keep exons found through OR/phase matching')

    parser.add_argument('--log', default = None, help = "file to write a log for debugging (optional)")

    args = parser.parse_args()

    if args.log:
        log_file = open(args.log,'w')
    else:
        log_file = open(os.devnull, 'w')

    if args.sep == "comma":
        delimiter = ','
    elif args.sep == "tab":
        delimiter = "\t"

    if "query_exon_info_table" in args:
        exon_info_dict = {}
        for line in open(args.query_exon_info_table):
            fields = line.replace('\n','').replace('\r','').split('\t')
            exon_id = fields[0] + ':' + fields[1]
            exon_info_dict[exon_id] = fields



    #Builds nested dictionaries for sorting by locus and co-ordinates
    locus_dict = {}

    for line in open(args.table):
        fields = line.replace('\r','').replace('\n','').split(delimiter)
        if not fields[1] in locus_dict:
            locus_dict[fields[1]] = {}
        locus_dict[fields[1]][int(fields[8])] = fields

    #Imports genome to evaluate splice sites and reading frames
    target_genome = genome.Genome(args.genome,truncate_names = True)

    #Here we go
    annotation_list = []
    for locus in locus_dict.keys():
        log_file.write('starting annotation of locus: ' + locus + '\n')
        working_annotation = []
        working_name = ["oops",0]
        #sets up variables needed for determining whether or not to start a new annotation
        last_from_start,last_from_end,last_strand,last_exon_num,last_num_exons,last_startphase,last_endphase,last_tstart,last_tend = 0,None,None,None,None,None,None,0,0
        #sorts coords
        sorted_coords = sorted(locus_dict[locus].keys())
        #goes through all hits on locus
        for start_coord_index in range(len(locus_dict[locus].keys()) + 1):
            #assigns values if not the first iteration (which is only used to get values for next iteration)
            if not start_coord_index == 0:
                qstart,qend,tstart,tend,startphase,endphase,qlen,exon_num,num_exons,strand,from_start,from_end = \
                    nextqstart,nextqstop,nexttstart,nexttstop,nextstartphase,nextendphase,nextqlen,nextexon_num,\
                    nextnum_exons,nextstrand,nextfrom_start,nextfrom_end
                hit = locus_dict[locus][sorted_coords[start_coord_index - 1]]
            #gets values for next iteration and to compare the comming hit to the current one and the previous one
            if start_coord_index < len(locus_dict[locus].keys()):
                next_start_coord = sorted_coords[start_coord_index]
                nexthit = locus_dict[locus][next_start_coord]
                nextqstart,nextqstop,nexttstart,nexttstop = int(nexthit[6]),int(nexthit[7]),int(nexthit[8]),int(nexthit[9])
                if nexttstart > nexttstop:
                    nextstrand = '-'
                else:
                    nextstrand = '+'
                nextstartphase,nextendphase,nextqlen,nextexon_num,nextnum_exons = int(nexthit[12]),\
                int(nexthit[13]), int(nexthit[14]) ,int(nexthit[15]),int(nexthit[16])
                if nextstrand == '-':
                    nextfrom_start = nextnum_exons - nextexon_num
                    nextfrom_end = nextexon_num
                else:
                    nextfrom_start = nextexon_num
                    nextfrom_end = nextnum_exons - nextexon_num
            #Has program move on if this is the first iteration (which is only used to get values for next iteration)
            if start_coord_index == 0:
                continue
            #Determines if exon is good and whether or not to start a new annotation

            if from_start == 0 and (nextfrom_start == 1 or working_annotation == [] or last_from_start > nextfrom_start or last_strand != nextstrand == strand):
                #found first/last exon! Starting new annotation
                if len(working_annotation) != 0:
                    #last annotation had something! Bulding an incomplete gene model for it (this can be scrapped later, after all)
                    last_annotation_almost_complete(last_strand,working_name,last_exon_num,last_tend,last_tstart,tstart,
                                    tend,log_file,args,exon_info_dict,target_genome,locus,working_annotation,
                                    last_startphase,last_num_exons,args.full_pseudoexon_search)
                working_name = [hit[0] +"coords" + locus + "-" + str(tstart),float(hit[11])]
                log_file.write('found start/end of gene, beggining new annotation near ' + str(tstart) + ' with working name "' + working_name[0] + '\n')
                if strand == "+":
                    working_annotation = exon_finder(min((tstart,tend)),max((tstart,tend)),strand,qstart,qend,
                        qlen,startphase,endphase,target_genome.genome_sequence,locus,is_start = True,
                        cluster = working_name[0].split('coord')[0],exon_number = exon_num, log_file = log_file,
                        full_pseudoexon_search = args.full_pseudoexon_search, exon_info_dict = exon_info_dict)
                elif strand == '-':
                    working_annotation = exon_finder(min((tstart,tend)),max((tstart,tend)),strand,qstart,qend,
                        qlen,startphase,endphase,target_genome.genome_sequence,locus,is_stop = True,
                        cluster = working_name[0].split('coord')[0],exon_number = exon_num, log_file = log_file,
                        full_pseudoexon_search = args.full_pseudoexon_search, exon_info_dict = exon_info_dict)
        
            elif (from_start == 1 and (working_annotation == [] or tstart - last_tstart > args.maxintron or
                                    (nextstrand != last_strand == strand or last_from_start > nextfrom_start > from_start or nextnum_exons == num_exons != last_num_exons))
                ) or last_from_start > nextfrom_start > from_start or last_strand != nextstrand == strand or tstart - last_tstart > args.maxintron:
                #looks like we've hit a new gene even though we didn't get the first exon. Starting new annotation
                if len(working_annotation) != 0:
                    #last annotation had something! Bulding an incomplete gene model for it (this can be scrapped later, after all)
                    last_annotation_almost_complete(last_strand,working_name,last_exon_num,last_tend,last_tstart,tstart,
                                    tend,log_file,args,exon_info_dict,target_genome,locus,working_annotation,
                                    last_startphase,last_num_exons,args.full_pseudoexon_search)
                working_name = [hit[0] +"coords" + locus + "-" + str(tstart),float(hit[11])]
                working_annotation = []
                log_file.write('found almost start/end of gene, beggining new annotation near ' + str(tstart) + ' with working name "' + working_name[0] + '\n')
                #OK, now trying to recover first/last exon
                if strand:
                    if strand == '-':
                        working_annotation, recovered = recover_missing_exon(strand,working_name[0].split('coord')[0],range(exon_num + 1,num_exons + 1),
                                                        False,True, [max((min((tend,tstart)) - 2000,max((last_tend,last_tstart)))),
                                                        min((tend,tstart))],log_file, args,exon_info_dict,target_genome,locus,tstart,last_strand,
                                                        working_annotation,args.full_pseudoexon_search)
                    elif strand == '+':
                        working_annotation, recovered = recover_missing_exon(strand,working_name[0].split('coord')[0],range(exon_num),
                                                        True,False, [max((min((tend,tstart)) - 2000,max((last_tend,last_tstart)))),
                                                        min((tend,tstart))],log_file,args,exon_info_dict,target_genome,locus,tstart,last_strand,
                                                        working_annotation,args.full_pseudoexon_search)
                else:
                    recovered = False
                #Now adding found exon
                if strand == "+" and working_annotation == []:
                    working_annotation.extend(exon_finder(min((tstart,tend)),max((tstart,tend)),strand,qstart,
                        qend,qlen,0,endphase,target_genome.genome_sequence,locus,is_start = True,nevermind_atg = True,
                        cluster = working_name[0].split('coord')[0],exon_number = exon_num, log_file = log_file,
                        full_pseudoexon_search = args.full_pseudoexon_search, exon_info_dict = exon_info_dict))
                    if_fail = "N"
                else:
                    working_annotation.extend(exon_finder(min((tstart,tend)),max((tstart,tend)),strand,qstart,
                        qend,qlen,startphase,endphase,target_genome.genome_sequence,locus,
                        cluster = working_name[0].split('coord')[0],exon_number = exon_num, log_file = log_file,
                        full_pseudoexon_search = args.full_pseudoexon_search, exon_info_dict = exon_info_dict))
                    if_fail = "C"
                if len(working_annotation) < 2:
                    if working_annotation[0][2]:
                        working_annotation[0][2] = working_annotation[0][2] + if_fail #Flags as incomplete because first/last exon was missed
                    else:
                        working_annotation[0][2] = if_fail
            ##Shouldn't need this since I'm starting a new annotation anytime the intron was too long
            # elif tstart - last_tstart > args.maxintron:
            #     #Intron was too long! Clearing annotation (or writing if almost done)
            #     if last_from_end == 1 and working_annotation != []:
            #         last_annotation_almost_complete(last_strand,working_name,last_exon_num,last_tend,last_tstart,tstart,
            #                        tend,log_file,args,exon_info_dict,target_genome,locus,working_annotation,
            #                        last_startphase,last_num_exons,args.full_pseudoexon_search)
            #     working_annotation = []
            #     working_name = ["oops",0]
            elif num_exons == last_num_exons and working_annotation != [] and strand == last_strand and from_start > last_from_start:
                #Found the next exon in the progression!
                if float(hit[11]) > working_name[1]:
                    working_name == [hit[0] +"coords" + locus + "-" + str(tstart),float(hit[11])]
                if from_start > last_from_start + 1:
                    #One or more exons were skipped! Attempting to recover
                    working_annotation, recovered = recover_missing_exon(strand,working_name[0].split('coord')[0],range(min([last_exon_num + 1,exon_num + 1]),
                                                                        max([last_exon_num ,exon_num])),False,False,[max((last_tend,last_tstart)),
                                                                        min((tstart,tend))],log_file,args,exon_info_dict,target_genome,locus,tstart,last_strand,
                                                                        working_annotation,args.full_pseudoexon_search)
                if from_end != 0:
                    working_annotation.extend(exon_finder(min((tstart,tend)),max((tstart,tend)),strand,qstart,qend,
                        qlen,startphase,endphase,target_genome.genome_sequence,locus,
                        cluster = working_name[0].split('coord')[0],exon_number = exon_num, log_file = log_file,
                        full_pseudoexon_search = args.full_pseudoexon_search, exon_info_dict = exon_info_dict))
                    if from_start > last_from_start + 1 and recovered != True:
                        #need to adjust exon start to account for mis-matched phase
                        if working_annotation[-1][2]:
                            working_annotation[-1][2] = working_annotation[-1][2] + "I"
                        else:
                            working_annotation[-1][2] = "I"
                        if not recovered:
                            if strand == "+":
                                exon_start_adjust = (last_endphase - startphase) % 3
                                working_annotation[-1][0] = working_annotation[-1][0] + exon_start_adjust
                                log_file.write('adjusting phase for missing exon, last end-phase = ' + str(last_endphase) + ', next start-phase =' +
                                            str(startphase) + ', exon-adjust = ' + str(exon_start_adjust) + '\n')
                            elif strand == "-":
                                exon_start_adjust = (endphase - last_startphase) % 3
                                working_annotation[-2][1] = working_annotation[-2][1] - exon_start_adjust
                                log_file.write('adjusting phase for missing exon, last end-phase = ' + str(last_endphase) + ', next start-phase =' +
                                            str(startphase) + ', exon-adjust = ' + str(exon_start_adjust) + '\n')
                else:
                    if strand == "+":
                        working_annotation.extend(exon_finder(min((tstart,tend)),max((tstart,tend)),strand,qstart,qend,
                            qlen,startphase,endphase,target_genome.genome_sequence,locus, is_stop = True,
                            cluster = working_name[0].split('coord')[0],exon_number = exon_num, log_file = log_file,
                            full_pseudoexon_search = args.full_pseudoexon_search, exon_info_dict = exon_info_dict))
                    elif strand == '-':
                        working_annotation.extend(exon_finder(min((tstart,tend)),max((tstart,tend)),strand,qstart,qend,
                            qlen,startphase,endphase,target_genome.genome_sequence,locus, is_start = True,
                            cluster = working_name[0].split('coord')[0],exon_number = exon_num, log_file = log_file,
                            full_pseudoexon_search = args.full_pseudoexon_search, exon_info_dict = exon_info_dict))
                    if from_start > last_from_start + 1 and not recovered == True:
                        #need to adjust exon start to account for mis-matched phase
                        if working_annotation[-1][2]:
                            working_annotation[-1][2] = working_annotation[-1][2] + "I"
                        else:
                            working_annotation[-1][2] = "I"
                        if not recovered:
                            if strand == "+":
                                exon_start_adjust = (last_endphase - startphase) % 3
                                working_annotation[-1][0] = working_annotation[-1][0] + exon_start_adjust
                                log_file.write('adjusting phase for missing exon, last end-phase = ' + str(last_endphase) + ', next start-phase =' +
                                            str(startphase) + ', exon-adjust = ' + str(exon_start_adjust) + '\n')
                            elif strand == "-":
                                exon_start_adjust = (endphase - last_startphase) % 3
                                working_annotation[-2][1] = working_annotation[-2][1] - exon_start_adjust
                                log_file.write('adjusting phase for missing exon, last end-phase = ' + str(last_endphase) + ', next start-phase =' +
                                            str(startphase) + ', exon-adjust = ' + str(exon_start_adjust) + '\n')
                    print(exontuples2gff(working_annotation, strand,working_name[0],locus))
                    log_file.write('writing complete annotation "' + working_name[0] + '"\n')
                    working_annotation = []
            else:
                continue
            last_from_start,last_from_end,last_strand,last_exon_num,last_num_exons,last_startphase,last_endphase,last_tstart,last_tend = \
                from_start,from_end,strand,exon_num,num_exons,startphase,endphase,tstart,tend
        if len(working_annotation) != 0:
            #last annotation was almost there! Bulding an incomplete gene model for it (this can be scrapped later, after all)
            last_annotation_almost_complete(last_strand,working_name,last_exon_num,last_tend,last_tstart,tstart,
                                    tend,log_file,args,exon_info_dict,target_genome,locus,working_annotation,
                                    last_startphase,last_num_exons,args.full_pseudoexon_search)

if __name__ == "__main__":
    main()


#!/usr/bin/env python

import sys
import os
import shutil
import argparse
from HAPpy import genome_fork as genome
from HAPpy import toolbox_for_HAP
import subprocess
import ete3
import time
import copy
import threading
import shlex
import tempfile
from collections import defaultdict,deque
from intervaltree import IntervalTree,Interval
import pickle

def output_fastas(clusters, prot_seq_dict, search_mode, fasta_dir,ref_genome_list = None,annotations = None):
    """Uses clusters dictionary built by "build_clusters" to break input sequences into fastas\
    for individual clusters. If search_mode = "exons", builds fastas for each exon."""
    sys.stderr.write('writing fastas for ' + str(len(clusters)) + ' clusters\n')
    cluster_dict = {}
    lengths_dict = {}
    if search_mode == 'exons':
        ref_genomes = []
        for i in range(len(ref_genome_list)):
            ref_genomes.append(genome.Genome(ref_genome_list[i]))
            ref_genomes[-1].read_gff(annotations[i])
    for cluster_index in range(len(clusters)):
        cluster = clusters[cluster_index]
        exon_file_root = fasta_dir + "/cluster" + str(cluster_index)
        cluster_dict[cluster_index] = []
        exon_number = None
        exon_phases = None
        exon_lengths = None
        for seqname in cluster:
            transcript_id = "_".join(seqname.split("_")[1:]) # because mafft adds "{number}_" to the front of the seq names
            if search_mode == "fl":
                seq_out = open(exon_file_root + '.fa','a')
                seq_out.write(">" + transcript_id + "\n" + prot_seq_dict[transcript_id] + '\n')
                seq_out.close()
            elif search_mode == "exons":
                for ref_genome in ref_genomes:
                    try:
                        cluster_dict[cluster_index].append(ref_genome.annotations.transcript[transcript_id])
                    except KeyError:
                        continue
                    ##This doesn't really fit the new program logic but might get re-incorporated depending on how I
                    ##decide to handle exon-based candidate loci parsing
                    #seq_out = open(exon_file_root + 'fullLenForHMM.fa','a')
                    #seq_out.write(ref_genome.annotations.transcript[transcript_id].get_fasta(seq_type='protein') + '\n')
                    #seq_out.close()
                    if not exon_number:
                        exon_number = len(ref_genome.annotations.transcript[transcript_id].child_list)
                    else:
                        if len(ref_genome.annotations.transcript[transcript_id].child_list) != exon_number:
                            sys.stderr.write("uh oh, transcripts with varying number of exons in: cluster " + str(cluster_index) + ", specifically gene " + transcript_id)
                            continue
                    sortlist = []
                    for CDS_id in ref_genome.annotations.transcript[transcript_id].child_list:
                        sortlist.append((ref_genome.annotations.CDS[CDS_id].get_coords(),CDS_id))
                    sortlist.sort()
                    if ref_genome.annotations.transcript[transcript_id].strand == "-":
                        sortlist.reverse()
                    if not exon_phases:
                        exon_phases = []
                        exon_lengths = []
                        current_phase = [0,0]
                        for cds_info in sortlist:
                            current_phase[1] = (current_phase[0] + cds_info[0][1] - cds_info[0][0] + 1) % 3
                            exon_lengths.append((cds_info[0][1] - current_phase[0] - cds_info[0][0] + 1) / 3)
                            exon_phases.append(current_phase[:])
                            current_phase[0] = current_phase[1]
                    for CDS_index in range(len(sortlist)):
                        cds_out = open(exon_file_root + "exon" + str(CDS_index) + 'of' + str(exon_number - 1) + 'phases' + str(exon_phases[CDS_index][0]) + 'and' + str(exon_phases[CDS_index][1])+ '.fa','a')
                        CDS_id = sortlist[CDS_index][1]
                        phase_slice_offset = (3 - exon_phases[CDS_index][0]) % 3
                        cds_out.write(">" + transcript_id + "-" + str(CDS_index) + '\n' + genome.Sequence(ref_genome.annotations.CDS[CDS_id].get_seq()[phase_slice_offset:]).translate() + '\n')
                        cds_out.close()
                    lengths_dict[cluster_index] = exon_lengths[:]
                    
def main():
    if sys.argv[1] == 'output_fastas':
        clusters = pickle.load(open('_clusters.pkl','rb'))
        prot_seq_dict = pickle.load(open('_psd.pkl','rb'))
        if '[' in sys.argv[4]:
            ref_genomes = eval(sys.argv[4])
            annotations = eval(sys.argv[5])
        else:
            ref_genomes = sys.argv[4]
            annotatios = sys.argv[5]
        output_fastas(clusters,prot_seq_dict,sys.argv[2],sys.argv[3],ref_genomes, annotations)

if __name__ == "__main__":
    main()

    



#!/usr/bin/env python
import argparse
import os
parser = argparse.ArgumentParser(description='splits query name output by HAP.py and builds table required for ABCENTH')
parser.add_argument('--table',default = None, help = 'table output by HAP.py')
parser.add_argument('--hmm_dir',default = None, help = "director with all cluster hmms")
args = parser.parse_args()

if args.table:
    for line in open(args.table):
        fields = line.replace('\n','').replace('\r','').split('\t')
        cluster = fields[0].split('exon')[0]
        exon_number = fields[0].split('exon')[1].split('of')[0]
        number_of_exons = fields[0].split('of')[1].split('phases')[0]
        start_phase = fields[0].split('phases')[1].split('and')[0]
        end_phase = fields[0].split('and')[1].split('.')[0]
        aa_len = fields[12]
        print('\t'.join([cluster] + fields[1:12] + [start_phase,end_phase,aa_len,exon_number,number_of_exons]))

elif args.hmm_dir:
    for hmm_file in os.listdir(args.hmm_dir):
        if hmm_file[-4:] == ".hmm" and not "fullLenForHMM" in hmm_file:
            cluster = hmm_file.split('exon')[0]
            exon_number = hmm_file.split('exon')[1].split('of')[0]
            number_of_exons = hmm_file.split('of')[1].split('phases')[0]
            start_phase = hmm_file.split('phases')[1].split('and')[0]
            end_phase = hmm_file.split('and')[1].split('.')[0]
            aa_len = open(args.hmm_dir + "/" + hmm_file).read().split('\n')[2].split()[1].replace('\r','')
            print('\t'.join([cluster,exon_number,number_of_exons,start_phase,end_phase,aa_len,os.path.abspath(args.hmm_dir) + '/' + hmm_file]))



#!/usr/bin/env python

#Seperates candidate loci by cluster and annotates them with genewise

import subprocess
import os
import tempfile
from HAPpy import genome_fork as genome
from HAPpy import toolbox_for_HAP
import threading
import time

def thammerin_for_parse_loci(hmm_file,hmm_dir,thammerin_path,candidate_loci_orfs_name):
    subprocess.call(thammerin_path + " -p " + hmm_dir + "/" + hmm_file +
                              " --frames_in " + candidate_loci_orfs_name + ' > thammerin_canloc_' +
                              hmm_file + '.tab', shell = True)


def parse_loci(candidate_loci_file,candidate_loci_seqs,hmm_dir,output_dir,thammerin_path = "thammerin.py",
               genewisedb_path = "genewisedb",hmmconvert_path = "hmmconvert", augustus_path = 'augustus',
               augustus_profile_dir = None, file_key = ".hmm", annotator = 'genewise', threads = 1, augustus_cfg = " ",
               augustus_species = 'fly'):
    candidate_loci_orfs = tempfile.NamedTemporaryFile()
    orfs_file_written = False
    for hmm_file in os.listdir(hmm_dir):
        if file_key in hmm_file:
            subprocess.call("mkdir -p " + output_dir + '/' + hmm_file[:-4], shell = True)
            if not orfs_file_written:
                subprocess.call(thammerin_path + " -p " + hmm_dir + "/" + hmm_file +
                                                  " -n " + candidate_loci_file + " --frames_out "
                                                  + candidate_loci_orfs.name + ' > thammerin_canloc_' + hmm_file + '.tab',
                                                  shell = True)
                orfs_file_written = True
            else:
                threading.Thread(target = thammerin_for_parse_loci, args =
                                 [hmm_file,hmm_dir,thammerin_path,candidate_loci_orfs.name]).start()
                while threading.active_count() > 1 + threads / 3:
                    time.sleep(1)
    while threading.active_count() > 1:
        time.sleep(1)
    subprocess.call('cat thammerin_canloc_* > thammerin_allcanloc.tab', shell = True)
    thammerin_lines = open('thammerin_allcanloc.tab').read().split('\n')
    candidate_loci_orfs.close()
    while "" in thammerin_lines:
        thammerin_lines.remove("")
    thammerin_scores = {}
    thammerin_hit_dict = {}
    for hit in thammerin_lines:
        fields = hit.split('\t')
        class_hit = fields[0].split('_hmmFile_')[1].split('_len_')[0][:-4]
        loc_name = fields[1]
        score = float(fields[11])
        if loc_name in thammerin_scores:
            if score > thammerin_scores[loc_name][0]:
                thammerin_scores[loc_name] = [score,class_hit]
            if class_hit in thammerin_hit_dict[loc_name]:
                thammerin_hit_dict[loc_name][class_hit].append(hit)
            else:
                thammerin_hit_dict[loc_name][class_hit] = [hit]
        else:
            thammerin_hit_dict[loc_name] = {class_hit:[hit]}
            thammerin_scores[loc_name] = [score,class_hit]
    for locus in thammerin_scores:
        cluster = thammerin_scores[locus][1]
        class_fasta = open(output_dir + "/" + cluster + '/candidate_loci.fasta','a')
        class_fasta.write('>' + locus + '\n' + candidate_loci_seqs[locus] + '\n')
        class_fasta.close()
        #locus_fasta = open(output_dir + "/" + cluster + '/' + locus + ".fa",'w')
        #locus_fasta.write('>' + locus + '\n' + candidate_loci_seqs[locus] + '\n')
        #locus_fasta.close()
        #starts prepping hints
    for cluster_folder in os.listdir(output_dir):
        threading.Thread(target= annotate, args = [annotator,hmm_dir,cluster_folder,output_dir,
                                                   thammerin_hit_dict,genewisedb_path,augustus_path,
                                                   hmmconvert_path, augustus_profile_dir, augustus_cfg,
                                                   augustus_species]).start()
        while threading.active_count() >= 2 + threads:
            time.sleep(1)



def annotate(annotator,hmm_dir,cluster_folder,output_dir,thammerin_hit_dict,genewisedb_path,
             augustus_path,hmmconvert_path,augustus_profile_dir, augustus_cfg, augustus_species):
    hints_file = open(output_dir + "/" + cluster_folder+ '/augustus_hints.gff','a')
    loci_list = subprocess.check_output('grep ">" ' + output_dir + '/' + cluster_folder + '/candidate_loci.fasta',
                                        shell = True).split('\n')
    while "" in loci_list:
        loci_list.remove("")
    for raw_locus in loci_list:
        locus = raw_locus[1:]
        hit_annotation = genome.read_blast_csv("\n".join(thammerin_hit_dict[locus][cluster_folder]).replace('\t',','))
        hints = genome.write_gff(hit_annotation,'augustus hint exonpart P').split('\n')
        for hint_line_index in range(len(hints)):
            hint_line_fields = hints[hint_line_index].split('\t')
            if int(hint_line_fields[4]) - int(hint_line_fields[3]) > 20:
                hint_line_fields[3] = str(int(hint_line_fields[3]) + 10)
                hint_line_fields[4] = str(int(hint_line_fields[4]) - 10)
                hints[hint_line_index] = "\t".join(hint_line_fields)
        hints_file.write('\n'.join(hints) + '\n')
    subprocess.call(hmmconvert_path + ' -2 ' + hmm_dir + '/' + cluster_folder + ".hmm > " + output_dir + '/' +
                   cluster_folder + '/' + cluster_folder + '.hmm', shell = True)
    if 'genewise' in annotator:
        subprocess.call(genewisedb_path + ' -cut 10 -pseudo -pep -gff -hmmer ' + output_dir + '/' +
                        cluster_folder + '/' + cluster_folder + '.hmm ' +
                        output_dir + '/' + cluster_folder + '/candidate_loci.fasta > ' + output_dir +
                        '/' + cluster_folder + '/candidate_loci.genewise 2>> genewise_log.txt' ,shell = True)
        genewise_lines,genewise_fasta = toolbox_for_HAP.parse_genewise(output_dir + "/" + cluster_folder +
                                                                       '/candidate_loci.genewise',
                                                                       name_prefix = cluster_folder.replace('.lenAdded','').replace('.hmm',''))
        ggtf_file = open(output_dir + "/" + cluster_folder + '/genewise.gtf','w')
        ggtf_file.write(''.join(genewise_lines))
        ggtf_file.close()
        gfasta_file = open(output_dir + "/" + cluster_folder + '/genewise.pep','w')
        gfasta_file.write(''.join(genewise_fasta))
        gfasta_file.close()
        if 'augustus' in annotator:
            genewiselines = open(output_dir + '/' + cluster_folder + '/candidate_loci.genewise').read().split('\n')
            hints_file.write('\n'.join(toolbox_for_HAP.genewise2aughints(genewiselines)) + '\n')
    hints_file.close()
    if 'augustus' in annotator:
        if augustus_profile_dir:
            augprofstring = " --proteinprofile=" + augustus_profile_dir + '/' + cluster_folder.replace('.hmm.lenAdded','') + '.prfl '
        else:
            augprofstring = " "
        print("running " + augustus_path + augprofstring + augustus_cfg + " --species=" + augustus_species + \
                        " --hintsfile=" + output_dir + "/" + cluster_folder + \
                        '/augustus_hints.gff '+ output_dir + '/' + cluster_folder + '/candidate_loci.fasta')
        subprocess.call(augustus_path + augprofstring + augustus_cfg + " --species=" + augustus_species +
                        " --hintsfile=" + output_dir + "/" + cluster_folder +
                        '/augustus_hints.gff '+ output_dir + '/' + cluster_folder + '/candidate_loci.fasta >> ' +
                        output_dir + "/" + cluster_folder + '/augustus.out', shell = True)
        augustus_lines,augustus_fasta = toolbox_for_HAP.parse_augustus(output_dir + "/" + cluster_folder +
                                                                       '/augustus.out', name_prefix = cluster_folder.replace('lenAdded','').replace('.hmm',''))
        gtf_file = open(output_dir + "/" + cluster_folder + '/augustus.gtf','w')
        gtf_file.write(''.join(augustus_lines))
        gtf_file.close()
        fasta_file = open(output_dir + "/" + cluster_folder + '/augustus.pep','w')
        fasta_file.write(''.join(augustus_fasta))
        fasta_file.close()



#!/usr/bin/env python

#Wraps hmmsearch to give the user a crude ability to conduct an HMMER search of a protein query
#against a nucleotide database.

#One day, one glorious day, the HMMER developers will write a tool to do this correctly. If you're frustrated
#with this script, join with me to bug them to add this feature themselves. I hear rumors that HMMER4 will be out
#soon- would it be too much to hope for that it will contain this function?

import subprocess
from HAPpy import genome_fork as genome
import argparse
import os
import tempfile
import shlex
import sys


def main():
    parser = argparse.ArgumentParser(description='thammerin: Pronounced Tee hammerin\', so pronounced identically \
                                    to tHMMERn but does NOT infinge on the copyrights of HHMI (HMMER) or \
                                    NCBI (tblastn). \nWraps hmmsearch to give the user a crude ability to conduct \
                                    a HMMER search of a protein query against a nucleotide database. \n\nProgram\
                                    dependencies:python (obviously), hmmer suite v3, and the genome_fork library from HAP.py. \n\
                                    Nota bene: for a 200 Mbp genome, building a database of translated sequences takes \
                                    ca. 5 minutes and running hmmsearch against this database takes ca. 5 seconds. Therefore \
                                    if you are going to run a search on the same genome multiple times it will GREATLY improve \
                                    perfomance if you save the database from the first run with the "--frames_out" option \
                                    and use it in subsequent runs with the "--frames_in" option.')

    parser.add_argument('-n','--nucleotide_seqs',dest='target_nucdb',
                        help = 'Nucleotide sequences to be searched (in fasta format)')
    parser.add_argument('-p','--protein_hmm', dest = 'hmm_file',
                        help = 'Query protein hmms (created using the hmmbuild program)')
    parser.add_argument('-e','--iEvalue_threshold', dest = 'iEval_thresh', default = float('inf'), type=float,
                        help = 'threshold "i-Evalue" for inclusion in results')
    parser.add_argument('-s','--min_orf_size', dest = 'min_orf_size', default = 10,
                        help = 'minimum orf to search with HMM profile')
    parser.add_argument('-f','--hmmsearch_filepath',dest = 'hmmsearch_filepath', default = 'hmmsearch',
                        help = 'optional path to hmmsearch program if not in your PATH variable')
    # parser.add_argument('--out_format', dest = 'out_format', default = 'standard',
    #                     help = 'output format, currently defunct, may eventually accept any combination of hmm name, target name, percent identity,\
    #                     alignment length, mismatch number, null field (for compatibility with blast standard output), hmm start,\
    #                     hmm end, target start, target end, hsp evalue, hmmsearch score, standard (all of the preceeding fields in\
    #                     that order), number of idenitical positions, number of similar positions, number of gaps, \
    #                     percentage of similar positions, any specific string (in quotes), target frame, hmm match sequence,\
    #                     target match sequeence, hmm length, and target length (whole seq, not just frame)')
    parser.add_argument('--frames_out' , default = False, help = 'File to save the translated frames \
                        if you want to keep them (e.g. to use for a later run)')
    parser.add_argument('--frames_in', default = False, help = 'File with frames from a previous run (will not re-make frames)')
    parser.add_argument('--hmmsearch_log', default = False, help = 'Log file to write the full hmmsearch results to (optional)')
    parser.add_argument('--add_lengths', default = True,type = bool, help = 'Whether to add qlen and slen to stdout (default = True)')

    args=parser.parse_args()

    #sets up default program paths, overwritten by any program paths passes with -f or --program_filepaths
    hmmsearch = args.hmmsearch_filepath
    target_nucdb = genome.Genome(args.target_nucdb,truncate_names = True)

    #Gets ORFs from the genome and hmmers them

    if args.frames_in:
        frames_file = args.frames_in
    else:
        if args.frames_out:
            frame_fasta = open(args.frames_out,'w')
        else:
            frame_fasta = tempfile.NamedTemporaryFile('w')
        for seq_id in target_nucdb.genome_sequence:
            frameonef = genome.Sequence(target_nucdb.genome_sequence[seq_id])
            frameoner = genome.Sequence(target_nucdb.genome_sequence[seq_id]).reverse_compliment()
            frames = [frameonef,genome.Sequence(frameonef[1:]), genome.Sequence(frameonef[2:]),
                    frameoner, genome.Sequence(frameoner[1:]),genome.Sequence(frameoner[2:])]
            fasta_list = []
            for frame_num in (0,1,2,3,4,5):
                frame = frames[frame_num]
                if frame_num < 3:
                    frame_offset = frame_num
                else:
                    frame_offset = len(frameonef) - frame_num + 3
                orfs = frame.translate(trimX = False).split('*')
                last_orf_end = 0
                for orf in orfs:
                    orf_start = last_orf_end
                    last_orf_end = last_orf_end + 1 + len(orf)
                    if len(orf) > args.min_orf_size:
                        fasta_list.append('>' + seq_id + "_frameOffset-" + str(frame_offset) + "_orfStart-" + str(orf_start) +
                                        '\n' + orf)
            frame_fasta.write('\n'.join(fasta_list) + '\n')
        frame_fasta.flush()
        frames_file = frame_fasta.name

    output_tempfile = tempfile.NamedTemporaryFile()
    output_logfile = tempfile.NamedTemporaryFile()
    subprocess.call(shlex.split(hmmsearch + " --domtblout " + output_tempfile.name + " " + args.hmm_file + " " + frames_file ), stdout = output_logfile)
    hmmresults = open(output_tempfile.name)
    for line in hmmresults:
        if line[0] != "#":
            fields = line.split()
            if float(fields[12]) < args.iEval_thresh:
                tname = fields[0]
                frame_offset = int(tname.split('_frameOffset-')[1].split('_orfStart-')[0])
                orf_start = int(tname.split('_orfStart-')[1])
                if frame_offset < 3:
                    strand = "+"
                    start = orf_start * 3 + frame_offset + int(fields[17]) * 3 - 2
                    stop = orf_start * 3 + frame_offset + int(fields[18]) * 3
                else:
                    strand = "-"
                    start = frame_offset - 3 * orf_start - 3 * int(fields[17]) + 3
                    stop = frame_offset - 3 * orf_start - 3 * int(fields[18]) + 1
                if args.add_lengths:
                    fields[13] += "\t" + fields[5] + '\t' + fields[2]
                sys.stdout.write("\t".join([fields[3],tname.split('_frameOffset-')[0],fields[21],'.','.',strand,
                                fields[15],fields[16],str(start),str(stop),fields[12],fields[13]]) + '\n')
    if args.hmmsearch_log:
        output_logfile.flush()
        hmmsearch_log = open(args.hmmsearch_log,'w')
        for line in open(output_logfile):
            hmmsearch_log.write(line)
        hmmsearch_log.close()
    try:
        output_tempfile.close()
        output_logfile.close()
        frame_fasta.close()
    except:
        pass

if __name__ == "__main__":
    main()


#!/usr/bin/python
#MAGOT: a functional and simple library for genomic analysis
#Sean McKenzie and Nelson Salinas
#See license on github (https://github.com/biorover/MAGOT/blob/master/LICENSE)
#Comments, feature requests, and friendly encouragement can be offered via github (https://github.com/biorover/MAGOT/issues)

#Known issues:
#Currently UTRs are not handled well. Most ways to import gff3 with UTRs encoded by exon feature regions which don't
#overlap with CDS feature regions will result in the UTRs being lost (because exon features are usually summarily ignored).
#Additionally, even if you manage to import UTRs, most output will likely ignore them. Bug us about it on a github ticket
#and we may well fix it.
#
#Importing GFFs still quite slow, likely having to do with ID assignment. Will try to fix soon.


import copy
import io

import numpy

verbose = True

def tab2fasta(tab_file):
    tabs = ensure_file(tab_file)
    newlines = []
    for line in tabs:
        fields = line.replace('\n','').replace('\r','').split('\t')
        newlines.append('>'+fields[0]+'\n'+fields[1])
    return "\n".join(newlines)


def fasta2tab(fasta_file):
    fasta = ensure_file(fasta_file)
    newlines = []
    for line in fasta:
        if line[0] == '>':
            newlines.append(line[1:].replace('\n','').replace('\r','') + '\t')
        else:
            newlines[-1] = newlines[-1] + line.replace('\n','').replace('\r','')
    return '\n'.join(newlines)


def ensure_file(potential_file):
    """takes an input that can be either a file location, opened file, or string, and returns an opened file or file like object.
    Used to make downstream applications robust to different input types"""
    if potential_file == None:
        return None
    elif type(potential_file).__name__ == "file":
        return potential_file
    else:
        try:
            return open(potential_file)
        except IOError:
            return StringIO.StringIO(potential_file)


def read_to_string(potential_file):
    """Tries to read "potential_file" into string- first as file location,
    then file, then string."""
    try:
        output_string = open(potential_file).read().replace('\r','')
    except IOError:
        try:
            output_string = potential_file.read().replace('\r','')
        except AttributeError:
            output_string = potential_file.replace('\r','')
    return output_string

def starjunc2gff(starjunc, output = 'string'):
    """parses junction file from STAR RNAseq aligner and returns an apollo-style gff3. Output can be "string", "list", or "print"."""
    if output != 'print':
        outlines = []
    junc_file = ensure_file(starjunc)
    strands = ['.','+','-']
    counter = 1
    for line in junc_file:
        ID = "StarAlignment_" + str(counter)
        counter = counter + 1
        atts = line.split('\t')
        stop1 = int(atts[1]) - 1
        start1 = stop1 - 20
        start2 = int(atts[2]) + 1
        stop2 = start2 + 20
        gffline1 = '\t'.join([atts[0],'star','match',str(start1),str(stop2),atts[6], strands[int(atts[3])],'.','ID=' + ID])
        gffline2 = '\t'.join([atts[0],'star','match_part',str(start1),str(stop1),atts[6], strands[int(atts[3])],'.','ID=' + ID + '-part1;Parent=' + ID])
        gffline3 = '\t'.join([atts[0],'star','match_part',str(start2),str(stop2),atts[6], strands[int(atts[3])],'.','ID=' + ID + '-part2;Parent=' + ID])
        if output == 'print':
            print(gffline1)
            print(gffline2)
            print(gffline3)
        else:
            outlines.append(gffline1)
            outlines.append(gffline2)
            outlines.append(gffline3)
    if output == "string":
        return "\n".join(outlines)
    elif output == "list":
        return outlines
    elif output == "print":
        pass
    else:
        print("invalid output format")


def apollo2genome(apollo_gff):
    apollo_list = read_to_string(apollo_gff).split('>')
    gff3 = apollo_list[0]
    fasta = '>' + apollo_list[1]
    return Genome(fasta,gff3,annotation_format='gff3')


def vulgar2gff(vulgarlist, feature_types=['match','match_part'],source='exonerate'):
    """takes vulgar alignment list (e.g. vulgarstring.split() ) and outputs gff lines. For eventual use with a read_exonerate function"""
    #sets variables to be added to gff lines
    qname = vulgarlist[0] + '-against-' + vulgarlist[4]
    qstart = vulgarlist[1]
    qend = vulgarlist[2]
    qstrand = vulgarlist[3]
    tname = vulgarlist[4]
    tstart = vulgarlist[5]
    tend = vulgarlist[6]
    tstrand = vulgarlist[7]
    score = vulgarlist[8]
    vulgartrips = vulgarlist[9:]
    addfeat = False
    IDnum = 1
    if tstrand == "+":
        tposition = int(tstart) + 1
    else:
        tposition = int(tstart)
        tend = str(int(tend)+1)
    #makes top level gff line for hit
    gfflines = ["\t".join([tname,source,feature_types[0],str(tposition),tend,score,tstrand,'.','ID='+qname])]
    #iterates over matches within hit to make bottom level feature gff lines
    for i in range(len(vulgartrips)):
        field = vulgartrips[i]
        if i % 3 == 0:
            if field == 'M' or field == 'S' or field == 'G' or field == 'F':
                if not addfeat:
                    addfeat = True
                    line_to_add = [tname,source,feature_types[1]]
                    coords = [str(tposition)]
                else:
                    pass
            elif addfeat:
                gfflines.append('\t'.join(line_to_add+[str(min(coords)),str(max(coords)),score,tstrand,'.',
                                                       'ID='+qname+'_'+feature_types[1]+str(IDnum)+';Parent='+qname]))
                IDnum = IDnum + 1
                addfeat = False
            else:
                pass
        if i % 3 == 2:
            if tstrand == "+":
                tposition = tposition + int(field)
            elif tstrand == "-":
                tposition = tposition - int(field)
            if addfeat == True:
                if tstrand == '+':
                    coords.append(str(tposition-1))
                elif tstrand == '-':
                    coords.append(str(tposition+1))
    if addfeat:
        gfflines.append('\t'.join(line_to_add+[str(min(coords)),str(max(coords)),score,tstrand,'.',
                                                       'ID='+qname+'_'+feature_types[1]+str(IDnum)+';Parent='+qname]))
    return '\n'.join(gfflines)


def read_exonerate(exonerate_output,annotation_set_to_modify = None,feature_types = ['match','match_part']):
    #Reads an exonerate file with a little help from vulgar2gff.
    if annotation_set_to_modify == None:
        annotation_set = AnnotationSet()
    else:
        annotation_set = annotation_set_to_modify
    exonerate_lines = ensure_file(exonerate_output)
    gfflines = []
    IDdic = {}
    qname = ""
    tname = ""
    for original_line in exonerate_lines:
        line = original_line.replace('\r','').replace('\n','')
        if line[:16] == "         Query: ":
            qname = line[16:]
        elif line[:16] == "        Target: ":
            tname = line[16:].replace(':[revcomp]','').replace('[revcomp]','')
            if tname[-1] == " ":
                tname = tname[:-1]
        elif line[:16] == "         Model: ":
            exonerate_model = line.split()[1].split(':')[0]
        elif line[:8] == "vulgar: ":
            vulgar_line_list = line[8:].split()
            #trying to makesure IDs are unique
            vulgar_line_list[0] = qname
            vulgar_line_list[4] = tname
            ID = vulgar_line_list[0] + '-against-' + vulgar_line_list[4]
            if ID in IDdic:
                vulgar_line_list[0] = vulgar_line_list[0] + "-" + str(IDdic[ID])
                IDdic[ID] = IDdic[ID] + 1
            else:
                IDdic[ID] = 1
            gfflines.append(vulgar2gff(vulgar_line_list,feature_types = feature_types,source = exonerate_model))
    read_gff("\n".join(gfflines), annotation_set_to_modify = annotation_set)
    if annotation_set_to_modify == None:
        return annotation_set


def write_longform_gff(annotation_set,keep_UTR_features = False):
    """returns gff string formated for compatability with Apollo genome annotation """
    gfflines = []
    fields = ['seqid','source','feature_type','get_coords()[0]','get_coords()[1]','score','strand','phase']
    #adds matches to gff
    if 'match' in annotation_set.__dict__:
        for match in annotation_set.match:
            match_obj = annotation_set.match[match]
            newline_list = []
            for field in fields:
                try:
                    newline_list.append(str(eval('match_obj.' + field)))
                except:
                    newline_list.append('.')
            attribute_list = ['ID=' + match_obj.ID]
            for attribute in match_obj.__dict__:
                if not attribute in fields+['annotation_set','parent','child_list','ID']:
                    attribute_list.append(attribute + '=' + eval('match_obj.' + attribute))
            newline_list.append(';'.join(attribute_list))
            gfflines.append('\t'.join(newline_list))
            for match_part in match_obj.child_list:
                match_part_obj = annotation_set[match_part]
                newline_list = []
                for field in fields:
                    try:
                        newline_list.append(str(eval('match_part_obj.' + field)))
                    except:
                        newline_list.append('.')
                attribute_list = ['ID=' + match_part_obj.ID,'Parent=' + match_part_obj.parent]
                for attribute in match_part_obj.__dict__:
                    if not attribute in fields+['annotation_set','parent','child_list','ID','coords']:
                        attribute_list.append(attribute + '=' + eval('match_part_obj.' + attribute))
                newline_list.append(';'.join(attribute_list))
                gfflines.append('\t'.join(newline_list))
        #adds genes to gff
    if 'gene' in annotation_set.__dict__:
        for gene in annotation_set.gene:
            gene_obj = annotation_set.gene[gene]
            newline_list = []
            for field in fields:
                try:
                    newline_list.append(str(eval('gene_obj.' + field)))
                except:
                    newline_list.append('.')
            attribute_list = ['ID='+gene_obj.ID]
            for attribute in gene_obj.__dict__:
                if not attribute in fields+['annotation_set','parent','child_list','ID']:
                    attribute_list.append(attribute + '=' + eval('gene_obj.' + attribute))
            newline_list.append(';'.join(attribute_list))
            gfflines.append('\t'.join(newline_list))
            for gene_child in gene_obj.child_list:
                gene_child_obj = annotation_set[gene_child]
                if gene_child_obj.feature_type == 'transcript':
                    transcript_obj = gene_child_obj
                    newline_list = []
                    for field in fields:
                        try:
                            newline_list.append(str(eval('transcript_obj.'+field)).replace('transcript','mRNA'))
                        except:
                            newline_list.append('.')
                    attribute_list = ['ID='+transcript_obj.ID,'Parent='+transcript_obj.parent]
                    for attribute in gene_obj.__dict__:
                        if not attribute in fields+['annotation_set','parent','child_list','ID']:
                            attribute_list.append(attribute + '=' + eval('transcript_obj.' + attribute))
                    newline_list.append(';'.join(attribute_list))
                    gfflines.append('\t'.join(newline_list))
                    exondict = {}
                    CDS_UTR_dict = {}
                    for transcript_child in transcript_obj.child_list:
                        transcript_child_obj = annotation_set[transcript_child]
                        line_base_list = []
                        for field in fields:
                            try:
                                line_base_list.append(str(eval('transcript_child_obj.'+ field)))
                            except:
                                line_base_list.append('.')
                        exon_attributes = 'ID=' + transcript_child_obj.ID + '-exon;Parent=' + transcript_child_obj.parent
                        transcript_child_attribute_list = ['ID=' + transcript_child_obj.ID, 'Parent=' + transcript_child_obj.parent]
                        for attribute in transcript_child_obj.__dict__:
                            if not attribute in fields+['annotation_set','parent','child_list','ID','coords']:
                                transcript_child_attribute_list.append(attribute + '=' + eval('transcript_child_obj.' + attribute))
                        transcript_child_attributes = ';'.join(transcript_child_attribute_list)
                        exondict[transcript_child_obj.coords] = '\t'.join(line_base_list).replace('CDS','exon').replace('UTR','exon') + '\t' + exon_attributes
                        CDS_UTR_dict[transcript_child_obj.coords] = '\t'.join(line_base_list) +'\t' + transcript_child_attributes
                    exondict_list = list(exondict)
                    exondict_list.sort()
                    CDS_UTR_dict_list = list(CDS_UTR_dict)
                    CDS_UTR_dict_list.sort()
                    #merges adjacent exons from abbutting UTRs and CDSs and writes exon gff lines
                    for i in range(len(exondict_list) - 1):
                        if exondict_list[i][1] + 1 == exondict_list[i+1][0]:
                            del exondict[exondict_list[i]]
                            new_exon_list = exondict[exondict_list[i+1]].split('\t')
                            exondict[exondict_list[i+1]] = '\t'.join(new_exon_list[:3]+[str(exondict_list[i][0]),str(exondict_list[i+1][1])]+new_exon_list[5:])
                        else:
                            gfflines.append(exondict[exondict_list[i]])
                    gfflines.append(exondict[exondict_list[-1]])
                    for i in CDS_UTR_dict_list:
                        if CDS_UTR_dict[i].split('\t')[2] == 'CDS' or keep_UTR_features:
                            gfflines.append(CDS_UTR_dict[i])
                else:
                    print('ERROR: currently only accepts AnnotationSets with gene format CDS/UTR -> transcript -> gene')
                    break
    return '\n'.join(gfflines)

def write_gff(annotation_set, gff_format = "simple gff3"):
    gff_lines = []
    for attribute in annotation_set.__dict__:
        if type(annotation_set.__dict__[attribute]) == dict:
            if len(annotation_set.__dict__[attribute]) > 0:
                if annotation_set.__dict__[attribute][list(annotation_set.__dict__[attribute])[0]].__class__.__name__ == "ParentAnnotation" or annotation_set.__dict__[attribute][list(annotation_set.__dict__[attribute])[0]].__class__.__name__ == "BaseAnnotation":
                    if annotation_set.__dict__[attribute][list(annotation_set.__dict__[attribute])[0]].parent == None:
                        for annotationID in annotation_set.__dict__[attribute]:
                            annotation = annotation_set.__dict__[attribute][annotationID]
                            gff_lines.append(annotation.get_gff(gff_format))
    return '\n'.join(gff_lines)

def write_bed(annotation_set, columns = 12):
    bed_lines = []
    for attribute in annotation_set.__dict__:
        if type(annotation_set.__dict__[attribute]) == dict:
            if len(annotation_set.__dict__[attribute]) > 0:
                if annotation_set.__dict__[attribute][list(annotation_set.__dict__[attribute])[0]].__class__.__name__ == "ParentAnnotation" or annotation_set.__dict__[attribute][list(annotation_set.__dict__[attribute])[0]].__class__.__name__ == "BaseAnnotation":
                    if annotation_set.__dict__[attribute][list(annotation_set.__dict__[attribute])[0]].parent == None:
                        for annotationID in annotation_set.__dict__[attribute]:
                            annotation = annotation_set.__dict__[attribute][annotationID]
                            bed_lines.append(annotation.get_bed(columns=columns))
    return '\n'.join(bed_lines)



def read_bed(bed, annotation_set_to_modify = None, parents_hierarchy = ['region_set'], base_features = 'region'):
    bed_file = ensure_file(bed)
    ID_set = set()
    new_ID_dict = {}
    if annotation_set_to_modify:
        annotation_set = annotation_set_to_modify
        for attribute in annotation_set.__dict__:
            if type(annotation_set.__dict__[attribute]) == dict:
                for feature_id in annotation_set.__dict__[attribute]:
                    ID_set.add(feature_id)
                    new_ID_dict[feature_id] = 2
    else:
        annotation_set = AnnotationSet()
    possible_fields = ['seqid','start','stop','name','score','strand','thickStart','thickEnd','itemRgb']
    for line in bed_file:
        if '\t' in line:
            fields = line.split('\t')
        else:
            fields = line.split()
        if len(fields) < 3:
            print("Error: line with fewer than three columns. First three BED columns are manditory.")
            print(line)
            continue
        #Figures out ID
        if len(fields) > 3:
            ID = fields[3]
        else:
            ID = fields[0] + ':' + fields[1] + '-' + fields[2]
        bol = False
        while ID in ID_set:
            bol = True
            ID = ID + '-' + str(new_ID_dict[ID])
            new_ID_dict[ID] = new_ID_dict[ID] + 1
        if not bol:
            new_ID_dict[ID] = 2
        ID_set.add(ID)
        #create parents
        for parent_index in range(len(parents_hierarchy)):
            parent_type = parents_hierarchy[parent_index]
            if parent_index == 0:
                child_list = []
            else:
                child_list = [parent_ID]
            if parent_index == len(parents_hierarchy) - 1:
                parent = None
                parent_ID = ID
            else:
                if parent_index == len(parents_hierarchy) - 2:
                    parent = ID
                else:
                    parent = ID + '-' + parents_hierarchy[parent_index + 1]
                parent_ID = ID + '-' + parent_type
            annotation_set[ID] = ParentAnnotation(parent_ID, fields[0], parent_type, child_list = child_list, parent = None, strand = ".", annotation_set = None, other_attributes = {})





def read_gff(gff,annotation_set_to_modify = None, base_features = ['CDS','match_part','similarity','region'],features_to_ignore = ['exon'],
    gff_version = "auto", parents_hierarchy = [], features_to_replace = [], IDfield = "ID", parent_field = "Parent", presets = None):
    """Assumptions of file:
    Relatively consistent, nothing crazy like mixed gff3 and gff2

    If a parent's children are base features, they should NOT overlap. E.g. if your file has exons and CDS, only
    give CDS as base_features and give exons as features_to_ignore.

    Although parent feature types may be used multiply in hierarchies, any given base feature
    will only be present in one hierarchy

    If parent features are not represented by lines (e.g. minimal gff2 format or cegma), they
    can be inferred from defline attributes given in parents_hierarchy. It is expected that the order they
    are given is consistent with the hierarchy (e.g. ['mRNA','gene']). If an underscore is used in
    this defline, the word before the underscore is used as the parent feature type. The "Parent"
    defline attribute is reserved for the imediate parent of the feature- if not given, this will be assumed to
    be the lowest-level parent specified in defline.
    """
    replace_dict = {"\n":"","\r":""}
    presets_dict = {}
    presets_dict["augustus"] = "features_to_ignore = ['gene','transcript','stop_codon','terminal','internal','initial','intron',\
        'start_codon','single']\nparent_field = None\nparents_hierarchy = ['transcript_id','gene_id']\nIDfield = None\ngff_version=2"
    presets_dict["RepeatMasker"] = "IDfield = None\nparent_field = None\nparents_hierarchy = ['match']\nfeatures_to_replace = [['similarity','match_part']]"
    presets_dict["CEGMA"] = "features_to_replace = [['First','CDS'],['Internal','CDS'],['Terminal','CDS'],['Single','CDS']]\n"
    if presets in presets_dict:
        exec(presets_dict[presets])
    version = gff_version
    gff_file = ensure_file(gff)
    for feature in features_to_replace:
        replace_dict["\t" + feature[0] + "\t"] = "\t" + feature[1] + "\t"
    if annotation_set_to_modify == None:
        annotation_set = AnnotationSet()
    else:
        annotation_set = annotation_set_to_modify
    #this dictionary and this list helps generate names if features passed with identical ID fields
    generate_new_ID_dict = {}
    allready_present_IDs = set()
    if annotation_set_to_modify:
        for attribute in annotation_set.__dict__:
            if type(annotation_set.__dict__[attribute]) == dict:
                for feature_id in annotation_set.__dict__[attribute]:
                    allready_present_IDs.add(feature_id)
    #figures out which feature types are base annotations
    base_dict = {}
    #here we go folks
    for original_line in gff_file:
        if original_line[0] != "#" and original_line.count('\t') == 8:
            line = original_line
            for string in replace_dict:
                line = line.replace(string,replace_dict[string])
            fields = line.split('\t')
            #Tries to detect GFF type if no presets or if presets are vague
            #
            if version == "auto":
                if "=" in fields[8]:
                    version = 3
                elif " " in fields[8]:
                    version = 2
                    if not IDfield == None:
                        if not " " + IDfield + " " in " " + fields[8].replace(';',' ') and parents_hierarchy == []:
                            IDfield = None
                            parent_field = None
                            if "gene_id" in fields[8] and "transcript_id" in fields[8]:
                                parents_hierarchy = ['transcript_id','gene_id']
                            elif "gene_id" in fields[8]:
                                parents_hierarchy = ['transcript_id','gene_id']
                else:
                    version = None
                    parents_hierarchy = ['transcript_id','gene_id']
                    IDfield = None
                    parent_field = None
            #Sets up basic variables- same for all gff types
            #
            ID = None
            parent = None
            other_attributes = {}
            seqid = fields[0]
            other_attributes['source'] = fields[1]
            feature_type = fields[2]
            if feature_type in features_to_ignore:
                continue
            coords = [int(fields[3]),int(fields[4])]
            coords.sort()
            coords = tuple(coords)
            try:
                other_attributes['score'] = float(fields[5])
            except ValueError:
                pass
            strand = fields[6]
            if fields[7] in ['0','1','2']:
                other_attributes['phase'] = int(fields[7])
            defline_dict = {}
            #Parsing the defline
            #
            for defline_field in fields[8].split(';'):
                if defline_field != "":
                    if version == None:
                        defline_dict['gene_id'] = defline_field
                    elif version == 2:
                        if '"' in defline_field:
                            defline_dict[defline_field.split()[0]] = defline_field.split('"')[1]
                        else:
                            try: defline_dict[defline_field.split()[0]] = defline_field.split()[1]
                            except:
                                print(defline_field)
                                return None
                    elif version == 3:
                        defline_dict[defline_field.split('=')[0]] = defline_field.split('=')[1]
            #Tries to figure out parent
            if parent_field != None:
                if parent_field in defline_dict:
                    parent = defline_dict[parent_field]
            elif parents_hierarchy != []:
                for parent_type in parents_hierarchy:
                    if parent_type in defline_dict:
                        parent = defline_dict[parent_type]
                        break
            #Assigns ID
            if IDfield != None:
                ID = defline_dict[IDfield]
            elif parent != None:
                ID = parent + '-' + feature_type
            else:
                ID = seqid + '-' + feature_type + fields[3]
            #checks if ID already in annotation_set and adjusts ID if necessary
            if ID in allready_present_IDs:
                try:
                    generate_new_ID_dict[ID] = generate_new_ID_dict[ID] + 1
                    ID = ID + "-" + str(generate_new_ID_dict[ID])
                except KeyError:
                    generate_new_ID_dict[ID] = 2
                    ID = ID + '2'
            allready_present_IDs.add(ID)
            #If no parent assigned the first time, but one should be created, sets parent name
            if parent == None and parents_hierarchy != []:
                parent = ID + '-' + parents_hierarchy[0]
            #Creates parents if necessary, adds to parent if exists
            if parent != None:
                child_to_assign = ID
                for parent_feature_index in range(len(parents_hierarchy)):
                    parent_feature = parents_hierarchy[parent_feature_index]
                    if parent_feature in defline_dict:
                        parent_feature_ID = defline_dict[parent_feature]
                        parent_feature_type = parent_feature.split('_')[0]
                    #guesses a parent id and creates when not in defline dict
                    else:
                        parent_feature_type = parent_feature.split('_')[0]
                        if parents_hierarchy[-1] in defline_dict:
                            parent_feature_ID = defline_dict[parents_hierarchy[-1]] + "-" + parent_feature_type
                        else:
                            parent_feature_ID = ID + "-" + parent_feature_type
                        #resets parent if this should be parent
                        if parent_feature_index == 0:
                            parent = parent_feature_ID
                    parents_parent = None
                    if parent_feature_index != len(parents_hierarchy) - 1:
                        for parents_parent_feature in parents_hierarchy[parent_feature_index + 1:]:
                            if parents_parent_feature in defline_dict:
                                parents_parent = defline_dict[parents_parent_feature]
                    if not parent_feature_type in annotation_set.__dict__:
                        annotation_set.__dict__[parent_feature_type] = {}
                    if parent_feature_ID in annotation_set.__dict__[parent_feature_type]:
                        if not child_to_assign in annotation_set.__dict__[parent_feature_type][parent_feature_ID].child_list:
                            annotation_set.__dict__[parent_feature_type][parent_feature_ID].child_list.append(child_to_assign)
                    else:
                        annotation_set.__dict__[parent_feature_type][parent_feature_ID] = ParentAnnotation(parent_feature_ID, seqid,
                                                                                        parent_feature_type, child_list = [child_to_assign],
                                                                                        parent = parents_parent, strand = strand,
                                                                                        annotation_set = annotation_set, other_attributes = {'source':fields[1]})
                    child_to_assign = parent_feature_ID
                #In case of no parent in parent hierarchy:
                if len(parents_hierarchy) == 0:
                    try:
                        annotation_set[parent].child_list.append(ID)
                    except KeyError:
                        print("""It seems that this line has a parent attribute but that that parent doesn't have a line itself nor
                        does this line have a defline attribute that specifies a parent type. I'm afraid this function can't currently
                        deal with that.""")
                        print(ID)
                        print(parent)
                        return None
            #fills other_attributes from defline
            for defline_attribute in defline_dict:
                if not defline_attribute in [IDfield,parent_field]:
                    other_attributes[defline_attribute] = defline_dict[defline_attribute]
            #And now to create the feature!
            if not feature_type in annotation_set.__dict__:
                annotation_set.__dict__[feature_type] = {}
            if feature_type in base_features:
                annotation_set.__dict__[feature_type][ID] = BaseAnnotation( ID, seqid, coords, feature_type, parent,
                                                                           strand , other_attributes, annotation_set)
            else:
                child_list = []
                annotation_set.__dict__[feature_type][ID] = ParentAnnotation(ID, seqid, feature_type, child_list,
                                                                             parent, strand, annotation_set, other_attributes)
    if annotation_set_to_modify == None:
        return copy.deepcopy(annotation_set)


def read_cegma_gff(cegma_gff,annotation_set_to_modify = None):
    """reads gff produced by CEGMA and returns AnnotationSet populated by CEGMA predictions"""
    annotation_set = read_gff(cegma_gff,annotation_set_to_modify = annotation_set_to_modify, presets = "CEGMA")
    if annotation_set_to_modify == None:
        return annotation_set


def read_blast_csv(blast_csv,annotation_set_to_modify = None,hierarchy = ['match','match_part'], source = 'blast', find_truncated_locname = False):
    """Reads csv output from blast (-outfmt 10) into an AnnotationSet object. Currently does not string hits together because I'm
    biased towards working on genes in tandem arrays where stringing hits together is annoying. May add option in future."""
    #reads blast_csv from file location, file, or string
    blast_file = ensure_file(blast_csv)
    #checks if annotation_set is given and creates annotation_set if not
    if annotation_set_to_modify == None:
        annotation_set = AnnotationSet()
    else:
        annotation_set = annotation_set_to_modify
    id_generator_dict = {}
    feature_type = hierarchy[-1]
    create_parents_chain = hierarchy[:-1]
    create_parents_chain.reverse()
    if not feature_type in annotation_set.__dict__:
        setattr(annotation_set, feature_type, {})
    if find_truncated_locname:
        if annotation_set.genome == None:
            print('"warning: find_truncated_locname" was set to true, but annotation set has no associated genome object so this cannot be done')
            find_truncated_locname = False
        else:
            genome_seqids = annotation_set.genome.get_seqids()
    for whole_line in blast_file:
        line = whole_line.replace('\r','').replace('\n','')
        fields = line.split(',')
        if len(fields) > 8:
            seqid = fields[1]
            if find_truncated_locname:
                if not seqid in genome_seqids:
                    for genome_seqid in genome_seqids:
                        if seqid == genome_seqid.split()[0]:
                            seqid = genome_seqid
                            break
            tstart = int(fields[8])
            tend = int(fields[9])
            if tstart < tend:
                coords = (tstart,tend)
                strand = '+'
            else:
                coords = (tend,tstart)
                strand = '-'
            score = fields[11]
            IDbase = fields[0]
            if IDbase in eval('annotation_set.' + feature_type):
                ID = IDbase + '-' + str(id_generator_dict[IDbase])
                id_generator_dict[IDbase] = id_generator_dict[IDbase] + 1
                while ID in eval('annotation_set.' + feature_type):
                    ID = IDbase + '-' + str(id_generator_dict[IDbase])
                    id_generator_dict[IDbase] = id_generator_dict[IDbase] + 1
            else:
                ID = IDbase
                id_generator_dict[IDbase] = 1
            other_attributes = {}
            other_attributes['evalue'] = fields[10]
            other_attributes['score'] = score
            parent = ID + '-' + create_parents_chain[0]
            child_to_set = ID
            for parent_index in range(len(create_parents_chain)):
                parent_feature = create_parents_chain[parent_index]
                if not parent_feature in annotation_set.__dict__:
                    setattr(annotation_set, parent_feature, {})
                if parent_index != len(create_parents_chain) - 1:
                    parent_to_set = ID + '-' + create_parents_chain[parent_index + 1]
                else:
                    parent_to_set = None
                annotation_set.__dict__[parent_feature][ID + '-' + parent_feature] = ParentAnnotation(ID + '-' + parent_feature, seqid, parent_feature,
                                                                                                    [child_to_set], parent_to_set, strand,
                                                                                                    annotation_set, other_attributes = {})
                child_to_set = ID + '-' + parent_feature
            eval('annotation_set.' + feature_type)[ID] = BaseAnnotation(ID, seqid, coords, feature_type, parent, strand,
                                                                        other_attributes, annotation_set)


    if annotation_set_to_modify == None:
        return annotation_set


class Genotype(tuple):
    """individual genotype to populate a GenotypeDict"""
    def __init__(self, phased = False, phase_group = None, qual = None):
        self.phased = phased
        self.phase_group = phase_group
        self.qual = qual


class Variant():
    """individual variants from reference genome sequence. Can be SNPs or polymorphism"""
    def __init__(self, ref_allele, alt_alleles, genotypes = None, variant_set = None, vcf_header_index = None):
        self.variant_set = variant_set
        self.ref_allele = ref_allele
        self.alt_alleles = alt_alleles
        self.genotypes = genotypes


class VariantSet(dict):
    """set of variants from reference genome sequence. Can be SNPs or polymorphism"""
    def __init__(self, genome = None, vcf_headers = None):
        self.genome = genome
        self.vcf_headers = vcf_headers


class AnnotationSet():
    """A set of annotations of a single genome. Each feature type (e.g. gene, transcript, exon, etc.)
    is stored in it's own dictionary as Annotations with their ID as their key (see "Annotation" class).
    The AnnotationSet itself also functions losely as a dictionary, in that any feature can be returned
    by indexing the AnnotationSet with the ID as a key (e.g. my_annotation_set["my_feature_ID"])"""
    def __init__(self, genome = None):
        self.gene = {}
        self.transcript = {}
        self.CDS = {}
        self.UTR = {}
        self.genome = genome

    def __getitem__(self,item):
        all_dicts = {}
        for attribute in dir(self):
            if type(eval("self." + attribute)) == dict:
                try:
                    all_dicts[item] = eval("self." + attribute)[item]
                except:
                    pass
        return all_dicts[item]

    def read_gff(self, gff, *args, **kwargs):
        kwargs["annotation_set_to_modify"] = self
        read_gff(gff, *args, **kwargs)

    def get_seqid(self, seqid):
        seqid_annotation_set = AnnotationSet()
        for attribute in dir(self):
            if type(eval("self." + attribute)) == dict:
                setattr(seqid_annotation_set,attribute,{})
                for feature in eval('self.'+ attribute):
                    feature_obj = eval('self.' + attribute)[feature]
                    if feature_obj.seqid == seqid:
                        eval('seqid_annotation_set.' + feature_obj.feature_type)[feature] = feature_obj
        return seqid_annotation_set

    def get_all_seqids(self):
        seqid_list = []
        for attribute in dir(self):
            if type(eval("self." + attribute)) == dict:
                for feature in eval('self.'+ attribute):
                    seqid_list.append(eval('self.' + attribute)[feature].seqid)
        return list(set(seqid_list))

    def read_exonerate(self, exonerate_output):
        read_exonerate(exonerate_output,annotation_set_to_modify = self)

    def read_blast_csv(self, blast_csv, hierarchy = ['match','match_part'], source = 'blast', find_truncated_locname = False):
        read_blast_csv(blast_csv, annotation_set_to_modify = self, hierarchy = hierarchy, source = source, find_truncated_locname = find_truncated_locname)

    def read_cegma_gff(self, cegma_gff):
        read_cegma_gff(cegma_gff, annotation_set_to_modify = self)

    def get_fasta(self,feature,seq_type = "nucleotide",longest=False,genomic = False):
        fasta_list = []
        for annotation in eval('self.' + feature):
                fasta_list.append(eval('self.' + feature)[annotation].get_fasta(seq_type = seq_type,longest = longest, genomic = genomic))
        return "\n".join(fasta_list)



class BaseAnnotation():
    """Bottom-most level annotation on a genome, for example CDS, UTR, Match_part, etc. Anything that should have no children"""
    def __init__(self, ID, seqid, coords, feature_type, parent = None, strand = ".", other_attributes = {}, annotation_set = None):
        #Sets up most attributes
        self.ID = ID
        self.seqid = seqid
        self.coords = coords
        self.feature_type = feature_type
        self.annotation_set = annotation_set
        for attribute in other_attributes:
            setattr(self, attribute, other_attributes[attribute])
        self.parent = parent
        self.strand = strand

    def get_coords(self):
        return self.coords

    def get_seq(self):
        try:
            if self.strand == '+' or self.strand == '.':
                return Sequence(self.annotation_set.genome.genome_sequence[self.seqid][self.coords[0]-1:self.coords[1]])
            elif self.strand == '-':
                return Sequence(self.annotation_set.genome.genome_sequence[self.seqid][self.coords[0]-1:self.coords[1]]).reverse_compliment()
            else:
                print(self.ID + ' has invalid strand value "' + self.strand + '"')
        except:
            print("either base_annotation has not annotation_set, or annotation_set has no genome, or genome has no\
            genome sequence, or genome sequence has no matching seqid, or coords are out of range on that seqid")
            print(self.seqid)

    def get_gff(self, gff_format = "simple gff3"):
        if self.annotation_set != None:
            fields = ['seqid','source','feature_type','get_coords()[0]','get_coords()[1]','score','strand','phase']
            fields_list = []
            for field in fields:
                try:
                    fields_list.append(str(eval('self.' + field)))
                except AttributeError:
                    fields_list.append('.')
            if gff_format in ["simple gff3","extended gff3","apollo gff3"]:
                defline = 'ID=' + self.ID
                if self.parent != None:
                    defline = defline + ';Parent=' + self.parent
            if gff_format == "extended gff3":
                for attribute in self.__dict__:
                    if type(self.__dict__[attribute]).__name__ == 'str' and not attribute in ['ID','Parent','score','strand','seqid','feature_type','phase','source']:
                        defline = defline + ';' + attribute + '=' + self.__dict__[attribute]
            elif gff_format[:13] == "augustus hint":
                gff_format_fields = gff_format.split()
                fields_list[2] = gff_format_fields[2]
                defline = "src=" + gff_format_fields[3]
                if len(gff_format_fields) > 4:
                    defline = defline + ";pri=" + gff_format_fields[4]
            elif gff_format == "gtf":
                defline = 'transcript_id ' + self.parent + ';gene_id ' + self.annotation_set[self.parent].parent
            fields_list.append(defline)
            if gff_format == "apollo gff3" and fields_list[2] == "CDS":
                return '\t'.join(fields_list).replace('\tCDS\t','\texon\t').replace('ID=','ID=ExonOf') + "\n" + '\t'.join(fields_list)
            else:
                return '\t'.join(fields_list)



class ParentAnnotation():
    """Parent of any BaseAnnotation. Examples include genes and transcripts. Suggested hierarchy for genes is
    CDS (as BaseAnnotation) -> transcript -> gene."""
    def __init__(self, ID, seqid, feature_type, child_list = [], parent = None, strand = ".", annotation_set = None, other_attributes = {}):
        self.ID = ID
        self.seqid = seqid
        self.feature_type = feature_type
        self.child_list = copy.copy(child_list)
        self.parent = parent
        self.annotation_set = annotation_set
        self.strand = strand
        for attribute in other_attributes:
            setattr(self, attribute, other_attributes[attribute])

    def get_coords(self):
        if len(self.child_list) > 0 and self.annotation_set != None:
            coords_list = []
            for child in self.child_list:
                child_object = self.annotation_set[child]
                if child_object.__class__.__name__ == 'ParentAnnotation':
                    coords_list = coords_list + list(child_object.get_coords())
                elif child_object.__class__.__name__ == 'BaseAnnotation':
                    coords_list = coords_list + list(child_object.coords)
                else:
                    print("for some reason you have children in ParentAnnotation " + self.ID + " which are neither \
                    ParentAnnotation objects nor BaseAnnotation object. Get your act together")
            return (min(coords_list),max(coords_list))

    def get_fasta(self, seq_type = "nucleotide", longest=False, genomic = False, name_from = 'ID'):
        """Returns fasta of this annotation's sequence. If this feature has multiple subfeatures (e.g. this is a gene
        and it has multiple transcripts), the sequence of each subfeature will be an entry in the fasta string."""
        if genomic == True:
            if self.annotation_set.genome != None:
                return ">" + self.ID + '\n' + self.annotation_set.genome.genome_sequence[self.seqid][self.get_coords()[0] - 1:self.get_coords()[1]] + '\n'
        elif len(self.child_list) > 0 and self.annotation_set != None:
            if self.annotation_set.genome != None:
                fasta_list = []
                child_type = self.annotation_set[self.child_list[0]].__class__.__name__
                if child_type == 'BaseAnnotation':
                    seq_list = []
                    child_dict = {}
                    for child in self.child_list:
                        child_obj = self.annotation_set[child]
                        try:
                            child_dict[child_obj.coords] = child_obj.get_seq()
                        except AttributeError:
                            print("ParentAnnotation has both ParentAnnotation and BaseAnnotation children!")
                            print(self.ID)
                        strand = child_obj.strand
                    children_in_correct_order = list(child_dict)
                    children_in_correct_order.sort()
                    if strand == '-':
                        children_in_correct_order.reverse()
                    for child in children_in_correct_order:
                        seq_list.append(child_dict[child])
                    if seq_type == "nucleotide":
                        new_seq = Sequence("".join(seq_list))
                    elif seq_type == "protein":
                        new_seq = Sequence("".join(seq_list)).translate()
                    else:
                        print(seq_type + ' is not valid seq_type. Please specify "protein" or "nucleotide".')
                    fasta_list.append('>' + self.__dict__[name_from] + '\n' + new_seq)
                else:
                    for child in self.child_list:
                        try:
                            child_fasta = self.annotation_set[child].get_fasta(seq_type=seq_type, name_from = name_from)
                            if child_fasta != "":
                                fasta_list.append(child_fasta)
                        except AttributeError:
                            print("ParentAnnotation has both ParentAnnotation and BaseAnnotation children!")
                            print(self.ID)
                if longest == True:
                    seqlens = {}
                    for seq in fasta_list:
                        seqlens[len("".join(seq.split('\n')[1:]))] = seq
                    return seqlens[max(list(seqlens))]
                else:
                    try:
                        return '\n'.join(fasta_list)
                    except:
                        return ""
            else: return ""
        else: return ""

    def get_gff(self, gff_format = "simple gff3"):
        """presets currently include: "simple gff3", "extended gff3", "gtf", "apollo gff3", "maker hint", and "augustus hint". Will eventually include more by request"""
        if self.annotation_set != None:
            fields = ['seqid','source','feature_type','get_coords()[0]','get_coords()[1]','score','strand','phase']
            fields_list = []
            parent_line = True
            for field in fields:
                try:
                    fields_list.append(str(eval('self.' + field)))
                except AttributeError:
                    fields_list.append('.')
            if gff_format in ["simple gff3","extended gff3","apollo gff3"]:
                defline = 'ID=' + self.ID
                if self.parent != None:
                    defline = defline + ';Parent=' + self.parent
            if gff_format == "extended gff3":
                for attribute in self.__dict__:
                    if type(self.__dict__[attribute]).__name__ == 'str' and not attribute in ['ID','Parent','score','strand','seqid','feature_type','phase','source']:
                        defline = defline + ';' + attribute + '=' + self.__dict__[attribute]
            elif gff_format == "apollo gff3":
                if 'Name' in self.__dict__:
                    defline = defline + ';Name=' + self.Name
                elif 'name' in self.__dict__:
                    defline = defline + ';Name=' + self.name
                else:
                    defline = defline + ';Name=' + self.ID
            elif gff_format[:13] == "augustus hint":
                parent_line = False
            elif gff_format == 'gtf':
                parent_line = False
            if parent_line:
                fields_list.append(defline)
                lines_list = ['\t'.join(fields_list)]
            else:
                lines_list = []
            child_dict = {}
            for child in self.child_list:
                child_object = self.annotation_set[child]
                if child_object.get_coords() not in child_dict:
                    child_dict[child_object.get_coords()] = child_object.get_gff(gff_format)
                else:
                    child_dict[(child_object.get_coords()[0],child_object.get_coords()[1] + len(child_dict))] = child_object.get_gff(gff_format)
            child_coords = list(child_dict)
            child_coords.sort()
            for child_index in child_coords:
                lines_list.append(child_dict[child_index])
            if gff_format == "apollo gff3":
                lines_list = '\n'.join(lines_list).split('\n')
                for line in lines_list[:]:
                    if line.split('\t')[2] == "CDS":
                        lines_list.remove(line)
                        lines_list.append(line)
            return '\n'.join(lines_list)

    def get_bed(self, columns = 12):
        """If children are base annotations, returns a bed-format line with the number of columns specified. Otherwise,
        returns bed-format lines for each child"""
        if self.annotation_set != None:
            lines_list = []
            child_blocks = []
            for child in self.child_list:
                child_object = self.annotation_set[child]
                if child_object.__class__.__name__ == "ParentAnnotation":
                    lines_list.append(child_object.get_bed(columns = columns))
                elif child_object.__class__.__name__ == "BaseAnnotation":
                    child_blocks.append((child_object.get_coords()[0] - self.get_coords()[0],
                                         child_object.get_coords()[1] - child_object.get_coords()[0] + 1))
            if child_blocks != []:
                if lines_list != []:
                    print("Weirdo feature " + self.ID + ", it has both ParentAnnotation and BaseAnnotation children.\
                    I'll output the line for this feature and it's ParentAnnotation children, but\
                    something is probably wrong")
                child_blocks.sort()
                child_block_sizes = []
                child_block_starts = []
                for child_block in child_blocks:
                    child_block_starts.append(str(child_block[0]))
                    child_block_sizes.append(str(child_block[1]))
                potential_fields = ['seqid','get_coords()[0] - 1','get_coords()[1]','ID','score','strand','thickStart','thickEnd','itemRgb']
                fields = []
                for field_number in range(columns):
                    if field_number < 9:
                        try:
                            fields.append(str(eval("self." + potential_fields[field_number])))
                        except AttributeError:
                            fields.append('.')
                    elif field_number > 11:
                        print("Only 12 fields in bed format!")
                    elif field_number == 9:
                        fields.append(str(len(child_block_sizes)))
                    elif field_number == 10:
                        fields.append(','.join(child_block_sizes))
                    elif field_number == 11:
                        fields.append(','.join(child_block_starts))
                lines_list.append('\t'.join(fields))
            return "\n".join(lines_list)



class Sequence(str):
    """DNA sequence. Has methods allowing reverse complimenting,
        translating, etc."""
    def reverse_compliment(self):
        """returns reverse compliment of self"""
        new_sequence_list = []
        compliment_dict = {'a':'t','t':'a','g':'c','c':'g','A':'T','T':'A','G':'C','C':'G','n':'n','N':'N','-':'-'}
        for residue in self[::-1]:
            try:
                new_sequence_list.append(compliment_dict[residue])
            except KeyError:
                new_sequence_list.append('n')
        return Sequence(''.join(new_sequence_list))

    def translate(self,library = {'TTT':'F','TTC':'F','TTA':'L','TTG':'L','CTT':'L','CTC':'L','CTA':'L','CTG':'L',
                                  'ATT':'I','ATC':'I','ATA':'I','ATG':'M','GTT':'V','GTC':'V','GTA':'V','GTG':'V',
                                  'TCT':'S','TCC':'S','TCA':'S','TCG':'S','CCT':'P','CCC':'P','CCA':'P','CCG':'P',
                                  'ACT':'T','ACC':'T','ACA':'T','ACG':'T','GCT':'A','GCC':'A','GCA':'A','GCG':'A',
                                  'TAT':'Y','TAC':'Y','TAA':'*','TAG':'*','CAT':'H','CAC':'H','CAA':'Q','CAG':'Q',
                                  'AAT':'N','AAC':'N','AAA':'K','AAG':'K','GAT':'D','GAC':'D','GAA':'E','GAG':'E',
                                  'TGT':'C','TGC':'C','TGA':'*','TGG':'W','CGT':'R','CGC':'R','CGA':'R','CGG':'R',
                                  'AGT':'S','AGC':'S','AGA':'R','AGG':'R','GGT':'G','GGC':'G','GGA':'G','GGG':'G'},
                  frame = 0, strand = '+',trimX = True):
        triplet = ""
        newseq = ""
        if strand == '+':
            seq = self
        elif strand == '-':
            seq = self.reverse_compliment()
        if len(seq) > (2 + frame):
            for residue_position in range(frame, len(self)):
                triplet = triplet + seq[residue_position].upper()
                if (residue_position + frame) % 3 == 2:
                    try:
                        newseq = newseq + library[triplet]
                    except KeyError:
                        newseq = newseq + 'X'
                    triplet = ""
            if trimX:
                if newseq[0] == 'X':
                    newseq = newseq[1:]
            return newseq

    def get_orfs(self, longest = False, strand = 'both', from_atg = False):
        if len(self) > 3:
            orflist = []
            if longest:
                candidate_list = []
                longest_orf_len = 0
            for frame in [0,1,2]:
                for strand in ['-','+']:
                    translated_seq = self.translate(frame=frame,strand=strand)
                    if translated_seq:
                        translated_seq_list = translated_seq.split('*')
                        for orf in translated_seq_list:
                            if from_atg:
                                try:
                                    output_orf = 'M' + ''.join(orf.split('M')[1:])
                                except IndexError:
                                    continue
                            else:
                                output_orf = orf
                            if longest:
                                if len(output_orf) > longest_orf_len:
                                    candidate_list.append(output_orf)
                                    longest_orf_len = len(output_orf)
                            else:
                                orflist.append(output_orf)
            if longest:
                return candidate_list[-1]
            else:
                return orflist


class GenomeSequence(dict):
    """genome sequence class, currently takes input in multi-fasta format."""
    def __init__(self,genome_sequence = None, truncate_names = False):
        #reads input file location, file, or string
        sequence_file = ensure_file(genome_sequence)
        #breaks sequence into sequence blocks (contigs, scaffolds, or chromosomes), adds sequence
        #   from each block as dictionary entry into self with block name as key.
        if sequence_file != None:
            seq = ""
            seqname = ""
            for line in sequence_file:
                if line[0] == ">":
                    if truncate_names == True:
                        seqid = line[1:].replace('\r','').replace('\n','').split()[0]
                    else:
                        seqid = line[1:].replace('\r','').replace('\n','')
                    if seq != "":
                        self[seqname] = seq
                        seq = ""
                    seqname = seqid
                else:
                    seq = seq + line.replace('\r','').replace('\n','')
            if seq != "":
                self[seqname] = seq


class Genome():
    """genome class, which contains sequence and annotations. Annotations can be given as annotation_set object, gff3, cegma_gff,
    blast_csv, or exonerate_output (just set annotation_format)."""
    def __init__(self,genome_sequence = None, annotations = None, varients = None, annotation_format = 'annotation_set', truncate_names = False):
        if genome_sequence.__class__.__name__ == 'GenomeSequence' or genome_sequence == None:
            self.genome_sequence = genome_sequence
        else:
            self.genome_sequence = GenomeSequence(genome_sequence, truncate_names = truncate_names)
        if annotations != None:
            if annotations.__class__.__name__ == "AnotationSet" and annotation_format == 'annotation_set':
                self.annotations = annotations
                self.annotations.genome = self
            elif annotation_format == 'gff3':
                self.annotations = read_gff(annotations)
                self.annotations.genome = self
            elif annotation_format == 'cegma_gff':
                self.annotations = read_cegma_gff(annotations)
                self.annotations.genome = self
            elif annotation_format == 'blast_csv':
                self.annotations = read_blast_csv(annotations)
                self.annotations.genome = self
            elif annotation_format == 'exonerate_output':
                self.annotations = read_exonerate(annotations)
                self.annotations.genome = self
        else:
            self.annotations = annotations

    def get_scaffold_fasta(self, seqid):
        return '>' + seqid + '\n' + self.genome_sequence[seqid]

    def get_genome_fasta(self, remove_spaces = False):
        fasta_list = []
        for seqid in self.genome_sequence:
            if remove_spaces:
                fasta_header = seqid.split()[0]
            else:
                fasta_header = seqid
            fasta_list.append('>' + fasta_header + '\n' + self.genome_sequence[seqid])
        return "\n".join(fasta_list)

    def write_apollo_gff(self, seqid, suppress_fasta = False):
        if self.genome_sequence != None and self.annotations != None:
            try:
                apollo_gff = write_gff(self.annotations.get_seqid(seqid),gff_format = 'apollo gff3')
                if not suppress_fasta:
                    apollo_gff = apollo_gff + '\n' + self.get_scaffold_fasta(seqid)
                return apollo_gff
            except:
                if not suppress_fasta:
                    return self.get_scaffold_fasta(seqid)
                else:
                    return ""
        else:
            print("genome object is either missing genome_sequence or annotations")

    def get_seqids(self, from_annotations = False):
        seqid_list = []
        warning = False
        if self.genome_sequence != None:
            for seqid in self.genome_sequence:
                seqid_list.append(seqid)
        if self.annotations != None and from_annotations:
            for seqid in self.annotations.get_all_seqids():
                if seqid not in seqid_list:
                    seqid_list.append(seqid)
                    warning = True
        if warning:
            print("warning, some annotations possessed seqids not found in sequence dictionary")
        return seqid_list

    def read_exonerate(self, exonerate_output):
        if self.annotations != None:
            self.annotations.read_exonerate(exonerate_output)
        else:
            self.annotations = read_exonerate(exonerate_output)
            self.annotations.genome = self

    def read_blast_csv(self, blast_csv, hierarchy = ['match','match_part'], source = 'blast', find_truncated_locname = False):
        if self.annotations == None:
            self.annotations = AnnotationSet()
            self.annotations.genome = self
        self.annotations.read_blast_csv(blast_csv, hierarchy = hierarchy, source = source, find_truncated_locname = find_truncated_locname)

    def read_cegma_gff(self, cegma_gff):
        if self.annotations != None:
            self.annotations.read_cegma_gff(cegma_gff)
        else:
            self.annotations = read_cegma_gff(cegma_gff)
            self.annotations.genome = self

    def read_gff(self, gff, *args, **kwargs):
        if self.annotations != None:
            self.annotations.read_gff(gff, *args, **kwargs)
        else:
            self.annotations = read_gff(gff, *args, **kwargs)
            self.annotations.genome = self

    def read_vcf(self, vcf):
        self.variants = read_vcf(vcf)


class position_dic(dict):
    def __init__(self, genome_sequence, dtype=bool):
        for seqid in genome_sequence:
            self[seqid] = numpy.zeros(len(genome_sequence[seqid]),dtype=dtype)

    def fill_from_annotations(self, annotation_set, feature, fill_type = "coords",fill_with = "1"):
        """accepts "start" and "coords" for fill_type"""
        for annotation in eval("annotation_set." + feature):
            annotation_obj = eval("annotation_set." + feature)[annotation]
            seqid = annotation_obj.seqid
            coords = annotation_obj.get_coords()
            try:
                parent = annotation_obj.parent
            except:
                pass
            ID = annotation_obj.ID
            if fill_type == "coords":
                for position in range(coords[0] - 1,coords[1]):
                    self[seqid][position] = eval(fill_with)
            elif fill_type == "start":
                self[seqid][coords[0] - 1] = eval(fill_with)

    def count_from_annotations(self, annotation_set, feature):
        count_list = []
        for annotation in eval("annotation_set." + feature):
            annotation_obj = eval("annotation_set." + feature)[annotation]
            seqid = annotation_obj.seqid
            coords = annotation_obj.get_coords()
            featureID = annotation_obj.ID
            try:
                parent = annotation_obj.parent
            except:
                pass
            ID = annotation_obj.ID
            count0 = 0
            count1 = 0
            for position in range(coords[0] - 1,coords[1]):
                if self[seqid][position] == 0:
                    count0 = count0 + 1
                elif self[seqid][position] == 1:
                    count1 = count1 + 1
            count_list.append([featureID,count0,count1])
        return count_list


    def at_content(self, genome_sequence):
        for seqid in self:
            for position in range(len(self[seqid])):
                if genome_sequence[seqid][position] in "ATat":
                    self[seqid][position] = 1

    def sliding_window_calculate(self, window_size, window_jump = 1, operation = "sum", output = "dict",
                                 threshold = 1, seqs_to_exclude = []):
        """operation may be "sum", "average", or "set". Output may be "dict","coords", or "annotation_set"."""
        if output == "dict":
            new_dic = {}
        elif output == "annotation_set":
            annotation_set = AnnotationSet()
            annotation_set.region = {}
        elif output == "coords":
            coords_list = []
        for seqid in self:
            if len(self[seqid]) > window_size and not seqid in seqs_to_exclude:
                if output == "dict":
                    new_dic[seqid] = []
                window_in = False
                coords_list = []
                for position in range(len(self[seqid][:-1 * window_size]) / window_jump):
                    window_start = position * window_jump
                    if operation == "set":
                        new_dic[seqid].append(set(self[seqid][window_start:window_start + window_size]))
                        continue
                    window_sum = numpy.sum(self[seqid][window_start:window_start + window_size])
                    if operation == "sum":
                        value = window_sum
                    elif operation == "average":
                        value = window_sum * 1.0 / window_size
                    if output == "dict":
                        new_dic[seqid].append(value)
                    elif output == 'annotation_set':
                        if value >= threshold:
                            if window_in == False:
                                window_in = True
                                if len(coords_list) == 0:
                                    coords_list.append([1 + window_start])
                                elif position > coords_list[-1][1]:
                                    coords_list.append([1 + window_start])
                        else:
                            if window_in == True:
                                window_in = False
                                if len(coords_list[-1]) == 1:
                                    coords_list[-1].append(window_start + window_size)
                                else:
                                    coords_list[-1][1] = window_start + window_size
                    elif output == "coords":
                        if threshold[0] <= value <= thredshold[1]:
                            coords_list.append([seqid, 1 + window_start, window_start + window_size])
                    elif type(output) == file:
                        output.write(seqid + '\t' + str(value) + '\n')
                    if window_start % 10000000 == 0 and verbose:
                        print("processed " + seqid + " to position " + str(position))
                if output == "annotation_set":
                    if len(coords_list) > 0:
                        if len(coords_list[-1]) == 1:
                            coords_list[-1].append(len(self[seqid]))
                        for coords in coords_list:
                            ID = seqid + "-window" + str(coords[0])
                            annotation_set.region[ID] = BaseAnnotation(ID, seqid, tuple(coords), "region", annotation_set = annotation_set)
                if verbose:
                    print("processed " + seqid)
                    if output == 'annotation_set':
                        for region in annotation_set.region:
                            if annotation_set.region[region].seqid == seqid:
                                print(str(annotation_set.region[region].coords))
        if output == "dict":
            return new_dic
        elif output == "annotation_set":
            return copy.deepcopy(annotation_set)
        elif output == "coords":
            return coords_list



#!/usr/bin/env python

import argparse
from HAPpy import genome_fork as genome
import re, os, subprocess, tempfile, shlex, sys

def orf_finder(sequence,startphase,stopphase,strand,expected_aa_len,length_variance,
               search_coords = None, is_start = False, is_stop = False, 
               hmm_profile = None, evalue='0.05', genewise_on_fail = True):
    """Finds all putative exons matching a given expected length and intron phase profile and \
    containing an open reading frame"""
    search_seq = sequence
    if search_coords:
        search_seq = sequence[search_coords[0]:search_coords[1]]
    if strand == '+':
        search_seq = genome.Sequence(search_seq)
    elif strand == '-':
        search_seq = genome.Sequence(search_seq).reverse_compliment()
    match_len = expected_aa_len * 3 + ((3 - startphase) % 3) + stopphase
    phase_matches = []
    startstop = ['AG','G[TC]']
    if is_start:
        startstop[0] = 'ATG'
        match_len = match_len - 3
    if is_stop:
        startstop[1] = "T(AG|GA|AA)"
        match_len = match_len - 3
    for variance in range(length_variance + 1):
        for direction in (1,-1):
            if variance == 0 and direction == -1: # prevents double returns of 0 variance
                continue
            for match in re.finditer(startstop[0] + '.{' + str(match_len + variance * 3 * direction) +
                                     '}' + startstop[1], search_seq):
                phase_matches.append([match.start(), match.start() + match_len + variance * 3 *
                                      direction + len(startstop[0]) + len(startstop[1])])
    exon_coords = []
    for match_coords in phase_matches:
        start = match_coords[0] + len(startstop[0]) 
        stop = match_coords[1] - len(startstop[1])
        if stop - (start + ((3 - startphase) % 3)) > 2:
            if not "*" in genome.Sequence(search_seq[start + ((3 - startphase) % 3) : stop ]).translate():
                #NB: returns 1-based coords consistent with gff and blast coords
                exon_coords.append([start + 1,stop])
                if is_start:
                    exon_coords[-1][0] = exon_coords[-1][0] - 3
                if is_stop:
                    exon_coords[-1][1] = exon_coords[-1][1] + 3
                if strand == '-':
                    exon_coords[-1] = [len(search_seq) - exon_coords[-1][1] + 1,len(search_seq) - exon_coords[-1][0] + 1]
                if search_coords:
                    exon_coords[-1] = [exon_coords[-1][0] + search_coords[0],exon_coords[-1][1] + search_coords[0]]     
    if hmm_profile and len(exon_coords) > 0:
        exon_coords = hmmsearch(hmm_profile,exon_coords,sequence,strand,startphase,evalue=evalue)
    if exon_coords == [] and genewise_on_fail and search_coords[1] - search_coords[0] < 10000:
        if not search_coords:
            search_coords = [0,None]
        exon_coords = genewisesearch(sequence,startphase,stopphase,strand,
                        hmm_profile, search_coords = search_coords,seqname = None)
    elif exon_coords != []:
        exon_coords = [ exon_coords + [False] ]
    return exon_coords

def hmmsearch(hmm_profile,exon_coords,sequence,strand,startphase, evalue= "0.05"):
    orf_file = tempfile.NamedTemporaryFile('w')
    for coords_index in range(len(exon_coords)):
        nuc_seq = genome.Sequence(sequence[exon_coords[coords_index][0] - 1:exon_coords[coords_index][1]])
        if strand == '-':
            nuc_seq = nuc_seq.reverse_compliment()
        pep_seq = genome.Sequence(nuc_seq[((3 - startphase) % 3):]).translate()
        orf_file.write(">coords" + str(coords_index) + '\n' +
                       pep_seq + '\n' )
    orf_file.flush()
    hmmout = subprocess.check_output(shlex.split('hmmsearch --max -E ' + str(evalue) + ' ' + hmm_profile + ' ' + 
        orf_file.name)).decode('utf8').split('\n')
    found_hit = False
    for line in hmmout:
        if line[:2] == ">>":
            found_hit = True
            return exon_coords[int(line[9:].replace('\r',''))]
            break
    if not found_hit:
        return []

def genewisesearch(sequence,startphase,stopphase,strand,hmm_profile,
                   search_coords = [0,None],seqname = None, log_file = open(os.devnull, 'w')):
    """searches an hmm profile against a sequence (optionally within a designated sub region) to \
    find boundaries of an exon interupted by stop codons or frame shifts"""
    os.environ['WISECONFIGDIR'] = os.path.dirname(genome.__file__) + '/cfgFiles'
    seqfile = tempfile.NamedTemporaryFile('w')
    seqfile.write(">temp_" + str(search_coords[0]) + "-" + str(search_coords[1]) + "\n" + sequence[search_coords[0]:search_coords[1]] + "\n")
    seqfile.flush()
    modelfile = tempfile.NamedTemporaryFile('w')
    subprocess.run(shlex.split('hmmconvert -2 ' + hmm_profile),stdout=modelfile)
    modelfile.flush()
    try:
        gwout = subprocess.check_output(shlex.split('genewisedb -sum -gff -hmmer ' \
            + modelfile.name + ' ' + seqfile.name),stderr = log_file).decode('utf8').split('\n')
    except subprocess.CalledProcessError:
        log_file.write('warning: genewise failed on one gene, continuing\n')
        return []
    #sys.stderr.write("\n".join(gwout) + "\n")
    tcoords = [None,None]
    qcoords = []
    keeplines = 0
    for line in gwout:
        if line != "":
            if line[0] == "/" and not tcoords[1]:
                tcoords[0] = None
        if len(line) > 4:
            if line[:4] == 'Bits' and not tcoords[0]:
                tcoords[0] = 'primed'
            elif tcoords[0] and not tcoords[1]:
                fields = line.split()
                start,stop = int(fields[5]), int(fields[6])
                if (strand == '+' and start < stop ) or (strand == "-" and start > stop):
                    tcoords = [start,stop]
                    keeplines = 1 + int(fields[7])
            elif tcoords[1] and tcoords[0] and "\tcds\t" in line and keeplines > 0:
                fields = line.split('\t')
                if fields[6] == strand:
                    coords = [int(fields[3]) + search_coords[0],int(fields[4]) + search_coords[0]]
                    if qcoords != []:
                        if fields[6] == '-':
                            while coords[1] > qcoords[-1][0]:
                                coords[1] += 3
                                if coords[1] + 3 >= coords[0]:
                                    break
                        elif fields[6] == '+':
                            while coords[0] < qcoords[-1][1]:
                                coords[0] += 3
                                if coords[0] + 3 >= coords[1]:
                                    break
                    qcoords.append(sorted(coords))
                    #sys.stderr.write(str(sorted(coords))+ "\n")
                    keeplines -= 1
            elif tcoords[1] and tcoords[0] and keeplines == 0:
                break
    qcoords.sort()
    if qcoords != []:
        #sys.stderr.write(str(qcoords) + '\n' + str((3 - startphase) % 3) + '\n' + str(stopphase) + "\n" + strand + "\n")
        if strand == '+':
            qcoords[0][0] = qcoords[0][0] - (3 - startphase) % 3
            qcoords[-1][-1] = qcoords[-1][-1] + stopphase
        elif strand == '-':
            qcoords[0][0] = qcoords[0][0] - stopphase
            qcoords[-1][-1] = qcoords[-1][-1] + (3 - startphase) % 3
    if len(qcoords) > 1:
        log_file.write('anotherone bites the dust\t')
    log_file.write(str(seqname) + '\t' + str([k + ['P'] for k in qcoords]) + '\n')
    return [k + ['P'] for k in qcoords]

def main():
    parser = argparse.ArgumentParser(description='Finds orfs between splice sites (or start and stop sites) which match a given \
                                     starting and ending intron phase')
    
    parser.add_argument('--sequence', help = 'sequence, raw or in fasta format')
    parser.add_argument('--startphase', type = int, help = 'phase of preceding intron')
    parser.add_argument('--stopphase', type = int, help = 'phase of following intron')
    parser.add_argument('--expected_aa_length', type = int, help = 'expected length (in amino acids) of exon product')
    parser.add_argument('--length_variance', type = int, help = 'permitted variance in exon length (in amino acids)')
    parser.add_argument('--strand', help = 'strand of expected feature')
    parser.add_argument('--search_coords', default = None, help = 'subset of sequence within which to search')
    parser.add_argument('--is_start', default = False, type = bool, help = 'Is this exon the first exon- i.e. should it start with atg instead of a splice site')
    parser.add_argument('--is_stop', default = False, type = bool, help = 'Is this exon the last exon- i.e. should it start with a stop codon instead of a splice site')
    parser.add_argument('--hmm_profile', default = None, help = "If provided, FindOrfs will search orfs against the provided\
                        hmm profile and return the best hit")
    
    args = parser.parse_args()

    sequence = ""
    seqcounter = 0
    for line in open(args.sequence):
        if not line[0] == ">":
            sequence = sequence + line.replace('\r','').replace('\n','')
        else:
            seqcounter += 1
            if seqcounter > 1:
                print("multiple seqs in fasta- not yet supported")
                break    
    exon_coords = orf_finder(sequence,args.startphase, args.stopphase, args.strand,
                             args.expected_aa_length, args.length_variance, args.search_coords,
                             args.is_start, args.is_stop, hmm_profile = args.hmm_profile)
    for coords in exon_coords:
        print(coords)

if __name__ == "__main__":
    main()



#!/usr/bin/env python

import argparse

parser = argparse.ArgumentParser(description='Selects best non-overlapping hits from a homology search hit table \
                                 (e.g. from tblastn), based on score or evalue')

parser.add_argument('--table',help = 'table of hits')
parser.add_argument('--sep', default = 'tab', help = 'table field delimiter, accepts "comma" and "tab" (default = tab')
parser.add_argument('--sortby',default = 'evalue', help = 'Criterion for "best" hit, accepts "evalue" \
                    (assumed to be field 11, minimizes) or "score" (assumed to be field 12, maximizes)')
parser.add_argument('--input_type', default = "blast_table", help = 'input file type- accepts "blast_table" (default) or \
                    "gtf"')
parser.add_argument('--output', default = "non-overlapping", help = 'specify "non-overlapping" (default) or "overlapping"')

args = parser.parse_args()
if args.sep == "comma":
    delimiter = ','
elif args.sep == "tab":
    delimiter = "\t"

sort_list = []
if args.input_type == "gtf":
    sortby_field = 3
elif args.sortby == 'evalue':
    sortby_field = 10
elif ags.sortby == 'score':
    sortby_field = 11
else:
    raise 'incorrect argument for "sortbby"'
    exit()

if args.input_type == "blast_table":
    coord1,coord2 = 8,9
elif args.input_type == "gtf":
    coord1,coord2 = 3,4

for line in open(args.table):
    fields = line.replace('\r','').replace('\n','').split(delimiter)
    sort_list.append([float(fields[sortby_field]),fields])

sort_list.sort()
if args.sortby == 'score':
    sort_list.reverse()

chunksize = 100000
lines_list = []
coords_dict = {}

for hit in sort_list:
    fields = hit[1]
    start, stop = min(int(fields[coord1]), int(fields[coord2])),max(int(fields[coord1]), int(fields[coord2]))
    if args.input_type == "blast_table":
        locus = fields[1]
    elif args.input_type == "gtf":
        locus = fields[0]
    coords = (start,stop)
    start_key = int(start / chunksize)
    stop_key = int(stop / chunksize)
    keys_list = [(locus,start_key)]
    if start_key != stop_key:
        for i in range(stop_key - start_key):
            keys_list.append((locus,keys_list[-1][-1] + 1))
    keep = True
    for key in keys_list:
        if key in coords_dict:
            for occupied_coord in coords_dict[key]:
                if occupied_coord[0] <= start <= occupied_coord[1] or \
                occupied_coord[0] <= stop <= occupied_coord[1] or \
                start < occupied_coord[0] < stop:
                    keep = False
                    if args.output == "overlapping":
                        print(delimiter.join(fields))
                    break
    if keep:
        for key in keys_list:       
            if not key in coords_dict:
                coords_dict[key] = [coords]
            else:
                coords_dict[key].append(coords)
        lines_list.append(delimiter.join(fields))

if args.output == "non-overlapping":
    print("\n".join(lines_list))





#!/usr/bin/env python

#   Features to be added later:
#proper splice site modeler (perhaps optimizing splice site score vs. difference from expected length), maybe MAYBE taking
#into account non-conserved exon phase.

#   Known issues:
#Will return annotations with arbitrary missing exons, but can't handle two exons missing in a row

import argparse
import genome_fork as genome
from FindOrfs import orf_finder


parser = argparse.ArgumentParser(description='ABCENTH, Annotation Based on Conserved Exons Noticed Through Homology. Takes a table of \
                                 exon alignments with phase and length information and tries to build good annotations. \
                                 Currently assumes that hits DO NOT contain stop codons.')

parser.add_argument('--table', help = 'Table of exon alignments. Expected to be in tblastn standard tab delimited format \
                    (-outfmt 6; query,target,blah,blah,blah,blah,qstart,qstop,tstart,tstop,evalue,score) with additional \
                    fields for query exon start phase, query exon end phase, query exon peptide length, query exon number (from 0), and highest query exon number \
                    (e.g. if their are 3 exons, they would be number 0, 1, and 2, so highest query exon number would be 2). \
                    Assumed to contain only non-overlapping hits (use HitTabFilter.py to find best non-overlapping hits)')

parser.add_argument('--genome', help = 'Genome for which to generate annotations (needed to ID splice sites)')

parser.add_argument('--sep', default = 'tab', help = 'Hit table field delimiter, accepts "comma" and "tab". Default= "tab"')

parser.add_argument('--maxintron', type = int, default = 100000, help = 'Maximum lenth of introns. Default = 100000')

parser.add_argument('--query_exon_info_table', help = 'Information about each query exon (for finding missing exons)')

parser.add_argument('--log', default = None, help = "file to write a log for debugging (optional)")

args = parser.parse_args()


if args.log:
    log_file = open(args.log,'w')

if args.sep == "comma":
    delimiter = ','
elif args.sep == "tab":
    delimiter = "\t"

if "query_exon_info_table" in args:
    exon_info_dict = {}
    for line in open(args.query_exon_info_table):
        fields = line.replace('\n','').replace('\r','').split('\t')
        exon_id = fields[0] + ':' + fields[1]
        exon_info_dict[exon_id] = fields


#The heart of gold of this program- the function to define an exon based on the coords and target phase
def exon_finder(tstart,tend,strand,qstart,qend,qlen,qstartphase,qendphase,seqdict,seqname,
                max_offset = 30, is_start = False, is_stop = False, nevermind_atg = False):
    """"finds exons with ORFs based on the requested parameters. Note that it is expected that tstart < tend, \
    even for minus strand features, so these might need to be reversed if coming from say tblastn. \
    Also coords are expected as 1-based (as output from blast), and are \
    converted internally to 0 based. It is also expected that the hit itself doesn't contain stop codons."""
    start = None
    end = None
    pseudo = False
    if strand == "+":
        phasestart_offset = (3 - qstartphase) % 3
        phasestop_offset = qendphase
        start_match_offset, stop_match_offset = 3 * (qstart - 1),3 * (qlen - qend)
    elif strand == '-':
        phasestart_offset = qendphase
        phasestop_offset = (3 - qstartphase) % 3
        start_match_offset, stop_match_offset = 3 * (qlen - qend), 3 * qstart
    ideal_start = tstart - phasestart_offset - start_match_offset
    ideal_end = tend + stop_match_offset + phasestop_offset
    pseudo_start = tstart - phasestart_offset
    pseudo_end = tend + phasestop_offset
    gc_start, gc_end = None, None
    for offset in range(0,max_offset + 3,3):
        if start:
            break
        for direction in [1,-1]:
            test_start = ideal_start - offset * direction
            test_seq = genome.Sequence(seqdict[seqname][test_start-1 + phasestart_offset:tend])
            if strand == "-":
                test_seq = test_seq.reverse_compliment()
            if not test_seq.translate():
                continue
            elif is_stop and strand == "-":
                if ideal_start - 1 < test_start < pseudo_start:
                    pseudo_start = test_start
                lastcodon = seqdict[seqname][test_start - 1:test_start + 2]
                if lastcodon.upper() in ['TTA','TCA','CTA'] and test_seq.translate().count('*') == 1:
                    start = test_start
                    break
            elif not "*" in test_seq.translate():
                if ideal_start - 1 < test_start < pseudo_start:
                    pseudo_start = test_start
                if is_start and strand == '+':
                    if nevermind_atg:
                        start = test_start
                        break
                    else:
                        firstcodon = seqdict[seqname][test_start - 1:test_start + 2]
                        if firstcodon.upper() == "ATG":
                            start = test_start
                            break
                else:
                    splicesite = seqdict[seqname][test_start-3:test_start-1]
                    if (strand == '+' and splicesite.upper() == "AG") or (strand == '-' and splicesite.upper() == "AC"):
                        start = test_start
                        break
                    elif strand == '-' and splicesite.upper() == "GC" and not gc_start:
                        gc_start = test_start
    if not start:
        if gc_start:
            start = gc_start
        else:
            pseudo = "P"
            start = pseudo_start
    for offset in range(0,max_offset + 3,3):
        if end:
            break
        for direction in [-1,1]:
            test_end = ideal_end - offset * direction
            if test_end - start < 3:
                break
            test_seq = genome.Sequence(seqdict[seqname][start - 1 + phasestart_offset:test_end - phasestop_offset])
            if strand == "-":
                test_seq = test_seq.reverse_compliment()
            if not test_seq.translate():
                continue
            elif is_stop and strand == "+":
                if ideal_end + 1 > test_end > pseudo_end:
                    pseudo_end = test_end
                lastcodon = seqdict[seqname][test_end - 3:test_end]
                if lastcodon.upper() in ['TAA','TGA','TAG'] and test_seq.translate().count('*') == 1:
                    end = test_end
                    break
            elif not "*" in test_seq.translate() or (is_stop and not "*" in test_seq.translate()[:-1]):
                if ideal_end + 1 > test_end > pseudo_end:
                    pseudo_end = test_end
                if is_start and strand == '-':
                    if nevermind_atg:
                        end = test_end
                        break
                    else:
                        firstcodon = seqdict[seqname][test_end - 3:test_end]
                        if firstcodon.upper() == "CAT":
                            end = test_end
                            break
                else:
                    splicesite = seqdict[seqname][test_end:test_end + 2]
                    if (strand == '+' and splicesite.upper() == "GT") or (strand == '-' and splicesite.upper() == "CT"):
                        end = test_end
                        break
                    elif strand == "+" and splicesite.upper() == "GC" and not gc_end:
                        gc_end = test_end
    if not end:
        if gc_end:
            end = gc_end
        else:
            pseudo = "P"
            end = pseudo_end
    return [start,end,pseudo]

#exon_finder(4179626,4179670,"-",1,40,40,0,0,cra.genome_sequence,'Chromosome3') #debug>>> exon_finder(4181323,4181442,"-",1,40,40,2,2,cra.genome_sequence,'Chromosome3') #debug

def exontuples2gff(exontuple_list,strand,feature_name,locus_name):
    outlines = []
    problems = []
    is_pseudo = False
    for exontuple in exontuple_list:
        outlines.append('\t'.join([locus_name,'ABCENTH','CDS',str(min(exontuple[:2])),str(max(exontuple[:2])),'.',strand,
                                  '.','gene_id ' + feature_name + ';transcript_id ' + feature_name + '-RA']))
        if exontuple[2]:
            is_pseudo = True
            problems.extend(list(exontuple[2]))
    if is_pseudo:
        problems = "".join(list(set(problems)))
        return '\n'.join(outlines).replace(';',problems + ';').replace('-RA',problems + '-RA')
    else:
        return '\n'.join(outlines)


#Builds nested dictionaries for sorting by locus and co-ordinates
locus_dict = {}

for line in open(args.table):
    fields = line.replace('\r','').replace('\n','').split(delimiter)
    if not fields[1] in locus_dict:
        locus_dict[fields[1]] = {}
    locus_dict[fields[1]][int(fields[8])] = fields

#Imports genome to evaluate splice sites and reading frames
target_genome = genome.Genome(args.genome)

#Defines a few operations that will need to be executed in different parts of the following code. NOT independent functions!
def recover_missing_exon(strand,cluster,exon_number,is_start,is_stop, search_coords):
    if "query_exon_info_table" in args:
        missing_exon_info = exon_info_dict[str(cluster) + ':' + str(exon_number)]
        missing_coords = orf_finder(target_genome.genome_sequence[locus],int(missing_exon_info[3]),
                                             int(missing_exon_info[4]),strand, int(missing_exon_info[5]),5, search_coords,
                                             is_stop = is_stop,is_start = is_start,hmm_profile = missing_exon_info[6])
    if args.log:
        log_file.write(locus + '\t' + str(tstart) + '\t' + str(cluster) + '\t' + str(exon_number) + '\t' + str(search_coords) + '\t' + str(missing_coords) + '\n')
    if missing_coords != []:
        working_annotation.append(missing_coords + [False])
        return True
    else:
        return False


def last_annotation_almost_complete():
    if last_strand == '-':
        is_start, is_stop, if_missing = True,False,"N"
    elif last_strand == '+':
        is_start,is_stop, if_missing = False,True,"C"
    recovered = recover_missing_exon(last_strand,working_name[0].split('coord')[0],eval(str(last_exon_num) + last_strand + "1"),is_start,is_stop,[max((last_tend,last_tstart)),min(max((last_tend,last_tstart)) + 2000,min((tstart,tend)))])
    if not recovered:
        if last_strand == '-' and last_startphase != 0:
            working_annotation[-1][1] = working_annotation[-1][1] - ((3 - last_startphase) % 3)
        if working_annotation[-1][2]:
            working_annotation[-1][2] = working_annotation[-1][2] + if_missing
        else:
            working_annotation[-1][2] = if_missing
    print(exontuples2gff(working_annotation, last_strand,working_name[0],locus))


#Here we go
annotation_list = []
for locus in locus_dict.keys():
    working_annotation = []
    working_name = ["oops",0]
    #sets up variables needed for determining whether or not to start a new annotation
    last_from_start,last_from_end,last_strand,last_exon_num,last_num_exons,last_startphase,last_endphase,last_tstart,last_tend = None,None,None,None,None,None,None,0,0
    #sorts coords
    sorted_coords = sorted(locus_dict[locus].keys())
    #goes through all hits on locus
    for start_coord_index in range(len(locus_dict[locus].keys()) + 1):
        #assigns values if not the first iteration (which is only used to get values for next iteration)
        if not start_coord_index == 0:
            qstart,qend,tstart,tend,startphase,endphase,qlen,exon_num,num_exons,strand,from_start,from_end = \
                nextqstart,nextqstop,nexttstart,nexttstop,nextstartphase,nextendphase,nextqlen,nextexon_num,\
                nextnum_exons,nextstrand,nextfrom_start,nextfrom_end
            hit = locus_dict[locus][sorted_coords[start_coord_index - 1]]
        #gets values for next iteration and to compare the comming hit to the current one and the previous one
        if start_coord_index < len(locus_dict[locus].keys()):
            next_start_coord = sorted_coords[start_coord_index]
            nexthit = locus_dict[locus][next_start_coord]
            nextqstart,nextqstop,nexttstart,nexttstop = int(nexthit[6]),int(nexthit[7]),int(nexthit[8]),int(nexthit[9])
            if nexttstart > nexttstop:
                nextstrand = '-'
            else:
                nextstrand = '+'
            nextstartphase,nextendphase,nextqlen,nextexon_num,nextnum_exons = int(nexthit[12]),\
            int(nexthit[13]), int(nexthit[14]) ,int(nexthit[15]),int(nexthit[16])
            if nextstrand == '-':
                nextfrom_start = nextnum_exons - nextexon_num
                nextfrom_end = nextexon_num
            else:
                nextfrom_start = nextexon_num
                nextfrom_end = nextnum_exons - nextexon_num
        #Has program move on if this is the first iteration (which is only used to get values for next iteration)
        if start_coord_index == 0:
            continue
        #Determines if exon is good and whether or not to start a new annotation
        if from_start == 0 and (nextfrom_start == 1 or working_annotation == []):
            #found first/last exon! Starting new annotation
            if last_from_end == 1 and len(working_annotation) != 0:
                #last annotation was almost there! Bulding an incomplete gene model for it (this can be scrapped later, after all)
                last_annotation_almost_complete()
            working_name = [hit[0] +"coords" + locus + "-" + str(tstart),float(hit[11])]
            if strand == "+":
                working_annotation = [exon_finder(min((tstart,tend)),max((tstart,tend)),strand,qstart,qend,qlen,startphase,endphase,target_genome.genome_sequence,locus,is_start = True)]
            elif strand == '-':
                working_annotation = [exon_finder(min((tstart,tend)),max((tstart,tend)),strand,qstart,qend,qlen,startphase,endphase,target_genome.genome_sequence,locus,is_stop = True)]
        elif from_start == 1 and (working_annotation == [] or tstart - last_tstart > args.maxintron or
                                  (nextstrand != last_strand or not nextfrom_start - last_from_start in [1,2] or not nextnum_exons == last_num_exons)):
            #found second/second-to-last exon and it seems like the first/last one was missing. Starting new annotation
            if last_from_end == 1 and len(working_annotation) != 0:
                #last annotation was almost there! Bulding an incomplete gene model for it (this can be scrapped later, after all)
                last_annotation_almost_complete()
            working_name = [hit[0] +"coords" + locus + "-" + str(tstart),float(hit[11])]
            working_annotation = []
            #OK, now trying to recover first/last exon
            if strand:
                if strand == '-':
                    me_is_start, me_is_stop = False,True
                elif strand == '+':
                    me_is_start,me_is_stop =  True,False
                recovered = recover_missing_exon(strand,working_name[0].split('coord')[0],eval(str(exon_num) + strand + "-1"),me_is_start,me_is_stop, [max((min((tend,tstart)) - 2000,max((last_tend,last_tstart)))),min((tend,tstart))])
            else:
                recovered = False
            #Now adding found exon
            if strand == "+" and working_annotation == []:
                working_annotation.append(exon_finder(min((tstart,tend)),max((tstart,tend)),strand,qstart,qend,qlen,startphase,endphase,target_genome.genome_sequence,locus,is_start = True,nevermind_atg = True))
                if_fail = "N"
            else:
                working_annotation.append(exon_finder(min((tstart,tend)),max((tstart,tend)),strand,qstart,qend,qlen,startphase,endphase,target_genome.genome_sequence,locus))
                if_fail = "C"
            if len(working_annotation) < 2:
                if working_annotation[0][2]:
                    working_annotation[0][2] = working_annotation[0][2] + if_fail #Flags as pseudogene because first/last exon was missed
                else:
                    working_annotation[0][2] = if_fail
        elif tstart - last_tstart > args.maxintron:
            #Intron was too long! Clearing annotation (or writing if almost done)
            if last_from_end == 1 and working_annotation != []:
                last_annotation_almost_complete()
            working_annotation = []
            working_name = ["oops",0]
        elif num_exons == last_num_exons and working_annotation != [] and strand == last_strand and (from_start == last_from_start + 1 or (from_start == last_from_start + 2 and nextfrom_start != last_from_start + 1)):
            #Found the next exon in the progression!
            if float(hit[11]) > working_name[1]:
                working_name == [hit[0] +"coords" + locus + "-" + str(tstart),float(hit[11])]
            if from_start == last_from_start + 2:
                #An exon was skipped! Attempting to recover
                recovered = recover_missing_exon(strand,working_name[0].split('coord')[0],eval(str(exon_num) + strand + "-1"),False,False,[max((last_tend,last_tstart)),min((tstart,tend))])
            if from_end != 0:
                working_annotation.append(exon_finder(min((tstart,tend)),max((tstart,tend)),strand,qstart,qend,qlen,startphase,endphase,target_genome.genome_sequence,locus))
                if from_start == last_from_start + 2 and not recovered:
                    #need to adjust exon start to account for mis-matched phase
                    if working_annotation[-1][2]:
                        working_annotation[-1][2] = working_annotation[-1][2] + "I"
                    else:
                        working_annotation[-1][2] = "I"
                    if strand == "+":
                        exon_start_adjust = (last_endphase - startphase) % 3
                        working_annotation[-1][0] = working_annotation[-1][0] + exon_start_adjust
                    elif strand == "-":
                        exon_start_adjust = (endphase - last_startphase) % 3
                        working_annotation[-2][1] = working_annotation[-2][1] - exon_start_adjust
            else:
                if strand == "+":
                    working_annotation.append(exon_finder(min((tstart,tend)),max((tstart,tend)),strand,qstart,qend,qlen,startphase,endphase,target_genome.genome_sequence,locus, is_stop = True))
                elif strand == '-':
                    working_annotation.append(exon_finder(min((tstart,tend)),max((tstart,tend)),strand,qstart,qend,qlen,startphase,endphase,target_genome.genome_sequence,locus, is_start = True))
                if from_start == last_from_start + 2 and not recovered:
                    #need to adjust exon start to account for mis-matched phase
                    if working_annotation[-1][2]:
                        working_annotation[-1][2] = working_annotation[-1][2] + "I"
                    else:
                        working_annotation[-1][2] = "I"
                    if strand == "+":
                        exon_start_adjust = (last_endphase - startphase) % 3
                        working_annotation[-1][0] = working_annotation[-1][0] + exon_start_adjust
                    elif strand == "-":
                        exon_start_adjust = (endphase - last_startphase) % 3
                        working_annotation[-2][1] = working_annotation[-2][1] - exon_start_adjust
                print(exontuples2gff(working_annotation, strand,working_name[0],locus))
                working_annotation = []
        else:
            continue
        last_from_start,last_from_end,last_strand,last_exon_num,last_num_exons,last_startphase,last_endphase,last_tstart,last_tend = \
            from_start,from_end,strand,exon_num,num_exons,startphase,endphase,tstart,tend
    if last_from_end == 1 and len(working_annotation) != 0:
                #last annotation was almost there! Bulding an incomplete gene model for it (this can be scrapped later, after all)
                last_annotation_almost_complete()



#!/usr/bin/env python
#Toolbox for HAP.py

#import genome_fork as genome
import subprocess
import time
import re

def genewise2aughints(genewiselines, src="G"):
    """Finds gff lines in a list of genewise lines and converts them to augustus hint format"""
    outlines = []
    for line in genewiselines:
        if line.count('\t') > 7:
            fields = line.split('\t')
            if fields[6] == '-':
                fields[3],fields[4] = fields[4],fields[3]
            if fields[2] == 'cds':
                outlines.append('\t'.join(fields[:2] + ['exon'] + fields[3:8] + ['src=' + src]))
            elif fields[2] == 'intron':
                if fields[6] == '+':
                    dss = fields[3]
                    ass = fields[4]
                elif fields[6] == '-':
                    dss = fields[4]
                    ass = fields[3]
                outlines.append('\t'.join(fields[:2] + ['dss',dss,dss] + fields[5:8] + ['src=' + src]))
                outlines.append('\t'.join(fields[:2] + ['ass',ass,ass] + fields[5:8] + ['src=' + src]))
    return(outlines)
    

def parse_genewise(genewise_file, use_gff_adjust = True, name_prefix = ""):
    """Pulls out gff lines and fasta sequences from genewise output"""
    gff_lines = []
    fasta_lines = []
    fasta_part = False
    for line in open(genewise_file):
        if line.count('\t') > 7 and "\tcds\t" in line:
            fields = line.split('\t')
            fields[8] = name_prefix + fields[8]
            fields[2] = "CDS"
            if use_gff_adjust:
                offset = int(fields[0].split('_')[1].split('-')[0])
                new_locus = fields[0].split('_')[0]
                gff_lines.append(gff_adjust(["\t".join(fields)],offset = offset,new_locus = new_locus)[0])
            else:
                gff_lines.append("\t".join(fields))
        elif line[0] == '>' and not "Results for " in line:
            fasta_lines.append(">" + name_prefix + line[1:])
            fasta_part = True
        elif line[0] == '/':
            fasta_part = False
        elif fasta_part:
            fasta_lines.append(line)
    return (gff_lines, fasta_lines)

def parse_augustus(augustus_file,name_prefix = None, use_gff_adjust = True):
    gtf_lines = []
    fasta_lines = []
    fasta_add = False
    for line in open(augustus_file):
        if line.count('\t') > 7 and "\tCDS\t" in line:
            fields = line.split('\t')
            working_name = fields[8].split('"')[-2]
            if name_prefix:
                working_name = name_prefix + "_" + working_name
                fields[8] = 'gene_id ' + working_name + ';transcript_id ' + working_name + '-RA\n'
            if use_gff_adjust:
                offset = int(fields[0].split('_')[1].split('-')[0])
                new_locus = fields[0].split('_')[0]
                gtf_lines.append(gff_adjust(["\t".join(fields)],offset,new_locus)[0])
            else:
                gtf_lines.append("\t".join(fields))
        elif "# protein sequence = [" in line:
            if not ']' in line:
                fasta_add = True
            fasta_lines.append('>' + working_name + '\n' + line.split('[')[1].replace(']',''))
        elif ']' in line and fasta_add:
            fasta_add = False
            fasta_lines.append(line[2:].replace(']',''))
        elif fasta_add:
            fasta_lines.append(line[2:])
    return (gtf_lines, fasta_lines)


def gff_adjust(gff_list,offset,new_locus = None):
    """Adds offset to each coordinate in gff_list"""
    adjusted_lines = []
    for line in range(len(gff_list)):
        fields = gff_list[line].split('\t')
        start,stop = fields[3],fields[4]
        fields[3] = str(min([int(start) + offset,int(stop) + offset]))
        fields[4] = str(max([int(start) + offset,int(stop) + offset]))
        if new_locus:
            fields[0] = new_locus  
        adjusted_lines.append('\t'.join(fields))
    return(adjusted_lines)




from setuptools import setup, find_packages

setup(
    name='HAPpy-ABCENTH',
    version='1.0',
    packages=[
        'HAPpy',
        'ABCENTH'
    ],
    install_requires=[
        'ete3',
        'numpy',
        'intervaltree'
    ],
    entry_points = {
        'console_scripts': [
            'thammerin = HAPpy.thammerin:main',
            'ABCENTH = ABCENTH.__main__:main',
            'HAPpy = HAPpy.__main__:main'
        ]
    },
    include_package_data = True
)
